# -*- coding: iso-8859-1 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

##--------------------------------------#######
#                  Feuille                                  #
##--------------------------------------#######
#    WxGeometrie
#    Dynamic geometrie, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA



# Ce module contient essentiellement la classe Feuille. C'est la classe qui va accueillir tous les objets geometriques.
# Elle fait l'intermediaire entre les objets et le Panel ou s'affichent les objets.

from objets import *


class Liste_objets:
    """Cette classe est juste un conteneur pour les objets de la feuille, qui sont tous ses attributs,
sauf ceux precedes de __ (attributs reserves pour tous les objets de la classe).
Elle contient aussi tous les objets geometriques.
Certaines methodes standard (comme __setattr__) sont aussi interceptees ou redefinies."""

    def __init__(self, feuille):
        
        import objets as GEO_objets
        
        objets_geomq = [couple for couple in GEO_objets.__dict__.items() if isinstance(couple[1], types.ClassType) and issubclass(couple[1], GEO_objets.Objet)]
        for couple in objets_geomq:
            self.__dict__[couple[0]] = couple[1] # ajoute au dictionnaire courant les objets geometriques, et uniquement eux (pas tout LIB.py !)
            
        import math
        self.__dict__.update(math.__dict__) # ajoute au dictionnaire les fonctions mathematiques courantes
        
        self.__dict__.update({"__feuille__" : feuille, "__self__" : self})
        
        # afin de pouvoir travailler dans l'espace de nom de feuille.objets en disposant de tous les objets, mais aussi des classes du module GEO_objets et des fonctions du module math.

        Objet.__feuille__ = self.Objet.__feuille__ = feuille
        
        self.__dict__.update({"__noms_reserves__" : math.__dict__.keys() + [nom for nom, objet in objets_geomq]})


    def __setattr__(self, nom, valeur):
        """Note: objets.__setattr__("a", 3) <-> objets.a = 3
        Les objets d'une feuille, contrairement aux objets Python, ne peuvent pas etre redefinis comme ca...
        En general, quand on essaie d'attribuer un nom qui existe deja, ce n'est pas volontaire.
        Pour eviter les erreurs, on impose de detruire explicitement l'objet avant que le nom puisse etre redonne."""
        
        # QUESTION : faut-il intercepter les valeurs numeriques, pour les transformer en objets de type Variable() ?

        #Objet.__feuille__ = self.__feuille__
        #self.Objet.__feuille__ = self.__feuille__
        
        #if hasattr(valeur, "__feuille__"):
        #    valeur.__feuille__ = self.__feuille__
            
        if type(valeur) in (int, long, float, str): # u=3 cree une variable
            valeur = self.Variable(valeur)

        if type(valeur) == list and type(valeur[0]) == str and len(valeur) in (1, 3): # t=["Bonjour!"] cree un texte
            valeur = self.Texte(*valeur)
                

        if type(valeur) == tuple and len(valeur) == 2: # A=(1,2) cree un point.
            valeur = self.Point(valeur[0], valeur[1])
        
        if not isinstance(valeur, Objet):
            raise TypeError, "type d'objet incorrect"

        if valeur.nom(): # si l'objet est deja reference sur la feuille, on en fait une copie.
            valeur = valeur.copy() # ainsi, A = B est remplace par A = B.copy()
        
        try:    # on verifie que le nom n'est pas encore attribue
            self.__verifier_nom__(nom)
        except:
            # Pour certains types d'objets (les points, textes et variables), le nom peut etre deja attribue.
            # Par exemple, A=Point(1,2) est valide meme si A existe deja.
            # L'objet A ne sera pas ecrase, mais actualise.
            # Dans cet exemple, les coordonnees de A sont actualisees en (1,2)
            # (Autrement dit, A=(1,2) devient un alias de A(1,2)).
            # De meme, pour une variable u, u=3 est compris comme u(3), qui est la syntaxe 'orthodoxe'.
            # Bien sur, il faut en particulier que la valeur soit un objet de meme type.
            # (A = Variable(3) n'est pas valide si A est un point !)
            if self.__dict__.has_key(nom) and self.__dict__[nom].__class__ == valeur.__class__:
                objet = self.__dict__[nom]
                if isinstance(objet, self.Point):
                    objet(valeur.x, valeur.y)
                elif isinstance(objet, self.Variable):
                    objet(valeur)
                elif isinstance(objet, self.Texte):
                    objet.label(valeur.texte)
                    objet(valeur.x, valeur.y)
                else:
                    raise
                return # le nom ne doit pas pointer sur valeur, mais toujours sur l'objet.
            else:
                raise    
        
        self.__dict__.update({nom: valeur})
        valeur.refresh()
        
        
    def __call__(self, objets_caches = True):
        """Retourne la liste des objets geometriques.
        Le parametre objets_caches indique s'il faut retourner les objets caches."""
        # Note: objets.__call__() <-> objets()
        return [obj for obj in self.__dict__.values() if isinstance(obj, Objet) and (objets_caches or obj.style("visible"))]
        
    
    def __noms__(self):
        "Retourne les noms de tous les objets geometriques."
        return [nom for nom, obj in self.__dict__.items() if isinstance(obj, Objet)]
    
    def __getitem__(self, nom):
        # Note: objets.__getitem__(nom) <-> objets[nom]
        try:
            if nom.startswith("__"):
                raise KeyError
            else:
                return self.__dict__[nom]
        except KeyError:
            print "Objet introuvable sur la feuille : " + nom
            raise KeyError
    
    
    def __verifier_nom__(self, nom):
        if nom.startswith("__"):
            self.__feuille__.erreur('un nom ne peut commencer par "__".')
            
        if not re.match("[A-Za-z_][A-Za-z0-9_]*$", nom):
            self.__feuille__.erreur("'%s' n'est pas un nom d'objet valide." %nom)
               
        if self.__dict__.has_key(nom):
            self.__feuille__.erreur("ce nom est deja utilise.")

    def effacer(self):
        for key, obj in self.__dict__.items():
            if isinstance(obj, Objet):
                del self.__dict__[key]
        
    
    


class Historique_feuille:
    """Historique de la feuille.
    Permet d'enregistrer l'etat de la feuille a un instant donne, et de le restaurer ensuite."""
    
    def __init__(self, feuille):
        self.__feuille__ = feuille
        self.n = param.nbr_annulations
        self.etats = []
        #self.etats_annules = []
        #self.derniere_commande = ""
        self.archiver()
        
    def archiver(self, commande = ""):
        # avec la molette de la souris, on effectue une succession rapide de zooms
        # pour eviter que ca ne remplisse l'historique, on archive alors l'etat actuel a la place du precedent.
        if commande.startswith("zoomer(") and self.derniere_commande.startswith("zoomer("):
            self.etats[-1] = self.__feuille__.sauvegarder()#False)
        else:
            self.etats.append(self.__feuille__.sauvegarder())#False))
        self.etats_annules = []
        self.etats = self.etats[-self.n:]
        self.derniere_commande = commande
    
    def annuler(self):
        if len(self.etats) > 1:
            etat_actuel = self.etats.pop()
            self.etats_annules.append(etat_actuel)
            self.etats_annules = self.etats_annules[-self.n:]
            self.restaurer(self.etats[-1])
            self.__feuille__.message("Action annulee.")
        else:
            self.__feuille__.message("Impossible d'annuler.")



    def refaire(self):
        if self.etats_annules:
            etat = self.etats_annules.pop()
            self.etats.append(etat)
            self.etats = self.etats[-self.n:]
            self.restaurer(etat)
            self.__feuille__.message("Action restauree.")
        else:
            self.__feuille__.message("Impossible de restaurer.")

                
    def restaurer(self, txt):
        commandes = regexp("set_fenetre[(][-0-9., ]*[)]", txt, "'__fenetre__=[' + x[12:-1] + ']'").split("\n")
        #debug(commandes)
        self.__feuille__.objets.effacer()     
        exec(commandes[0], self.__feuille__.__dict__)
        for commande in commandes[1:]:
            exec(commande, self.__feuille__.objets.__dict__)
        self.__feuille__.refresh()








class Feuille:
    "Feuille de travail"

    def __init__(self, parent = None, canvas = None):
        self.__fenetre__ = param.fenetre
        self.objets = Liste_objets(self)
        self.historique = Historique_feuille(self)
        self.parent = parent
        if not canvas and parent:
            self.canvas = self.parent.canvas
        else:
            self.canvas = canvas
            
            
        #self.renomme(nom)   # enregistre la feuille dans self.parent.feuilles sous le nom "nom"
        #self.parent.actuelle = self # devient la feuille courante
        Objet.__feuille__ = self # les objets sont crees dans cette feuille par defaut
                
        
        


#########################################################################################    

# Les fonctions suivantes permettent de passer facilement de l'objet a son nom, et vice-versa...
# Note : il n'y a pas necessairement bijection entre l'ensemble des objets et celui des noms,
# deux noms pouvant correspondre au meme objet. Dans ce cas, le nom renvoye est un des noms de l'objet.


    def nom_objet(self, objet, liste = False):
        # Renvoie le nom de l'objet, ou celui de la feuille en l'absence d'objet.
        liste_noms = [couple[0] for couple in self.objets.__dict__.items() if couple[1] is objet]
        if liste: return liste_noms # si liste == True, renvoie la liste des noms.
        if liste_noms: return liste_noms[0] # sinon, le premier nom.
        return  ""

    def nom(self):
        return self.parent.nom_feuille(self)

    def objet(self, nom):    # retourne l'objet associe au nom "nom"
        return self.objets[nom]
        
    
#######################################################################################

# Methodes se rapportant a la feuille elle-meme

# A REECRIRE

    
#    def renomme(self, nom):
#        self.parent.feuilles[self] = nom # renomme la feuille
    
#    def supprimer(self):
#        self.parent.feuilles.pop(self) #
        
    def inventaire(self):
        if param.debug:
            for obj in self.objets():   print "- " + obj.nom() + " : " + repr(obj) + " (" + obj.type() + ")"
        liste = [obj.nom() + " est " + obj.titre() + (not obj.style("visible") and " (invisible)" or "") for obj in self.objets()]
        liste.sort()
        return liste
            
            
#######################################################################################
    
# Gestion de l'affichage

#    -> est largement gere desormais par le module MOD_affichage.py
    
    def message(self, messg): # A REECRIRE
        if self.canvas:
            messg = "Feuille %s - %s" %(self.nom(), messg)
            self.canvas.message(messg)
        else:
            print messg


        
#    def affiche(self):
#        for obj in self.objets():
#            try:
#                if not obj.nom().startswith("_"): obj.affiche()  # Les noms de la forme "_254" par exemple correspondent a des objets intermediaires utilises par la construction...
#            except:
#                self.message("Erreur lors de l'affichage de %s." % obj.nom())
#                if param.debug:
#                    print "Erreur lors de l'affichage de l'objet " + obj.nom()
#                    raise
        
        
        
    def refresh(self, *type):
        """A utiliser apres une commande annuler ou refaire (affichage lent).
        
        Si 'type' est precise, ne refraichit que les objets du ou des type(s) specifies.
        C'est utile en particulier pour rafraichir les droites.
        Quand on elargit la fenetre d'affichage, il faut en effet
        redessiner les droites pour donner une impression d'infini !"""
        
        for objet in self.objets():
            if not type or isinstance(objet, type):
                try:
                    objet.affiche()
                except:
                    self.message("Erreur lors de l'affichage de %s." % objet.nom())
                    if param.debug:
                        print "Erreur lors de l'affichage de l'objet " + objet.nom()
                        print_error()
        if self.parent:
            self.canvas.rafraichir_axes = True
            self.parent.affiche()
               
    
    
    def affiche(self):
        "Affichage rapide."
        for representation in [objet.__representation__ for objet in self.objets()]:
            self.canvas.graph.ajouter(representation)
        
        
        
#########################################################################################    


# Sauvegarde, exportation... du travail en cours.


        
    def sauvegarder(self): #, creer_feuille = True):
        "Cree un fichier python qui permettra de recreer la figure et l'ensemble des objets."
        
        objets = [obj for obj in self.objets() if not isinstance(obj, Cote)]
        objets.sort(key=lambda x:getattr(x,"hierarchie"))
        # on doit enregistrer les objets dans le bon ordre (suivant la hierarchie)
        #texte = creer_feuille and "creer_feuille()\n" or ""
        texte = "set_fenetre(" + str(self.fenetre())[1:-1] + ")\n"
        texte += "\n".join([objet.nom() + "=" + repr(objet) for objet in objets]) + "\n"
        #texte += "\n".join(["%s.style(**%s)" %(objet.nom(), objet.style()) for objet in objets])
        #texte += creer_feuille and "\nrafraichir_feuille()" or ""
        return texte    


        
        
        
#########################################################################################    

# Gestion du zoom et des coordonnees, reglage de la fenetre d'affichage.

#    -> est gere desormais par le module MOD_affichage.py


    def fenetre(self):
        return self.__fenetre__
        # retourne (xmin, xmax, ymin, ymax).

#########################################################################################    

# Diverses fonctionnalites de la feuille, utilisees par les objets.
# C'est un peu la boite a outils :-)




    def nom_alea(self, prefixe = "obj"): # utilise pour la generation d'objets de noms toujours differents
        #print self.objets.__noms__()
        liste = [int(s[len(prefixe):]) for s in self.objets.__noms__() if re.match(prefixe + "[0-9]+$", s)]
        #print liste
        return prefixe + str(max(liste or [0]) + 1)


    def erreur(self, message):
        self.message("Erreur : " + message)
        raise NameError, message

