# -*- coding: iso-8859-1 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

##--------------------------------------#######
#                   Objets                    #
##--------------------------------------#######
#    WxGeometrie
#    Dynamic geometrie, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA



from pylib import *
from mathlib import *

#################################################################################################################


# NOTE : les vassaux ne sont pas parfaitement bien geres :
#    Un objet ne doit pas etre a la fois vassal et suzerain d'un autre.
#    Une telle situation provoquerait un plantage lors de la suppression d'un des deux objets (boucle recursive infinie).
#    Ceci ne pose a priori pas de problemes, sauf un peu de gaspillage de memoire :
#    en effet, les objets geometriques intermediaires introduits pour construire un objet ne sont pas systematiquement detruits quand l'objet est supprime.


class Objet:
    """Un objet geometrique."""
    
    __feuille__ = None
    # NOTES:
    # Apres l'importation du module, on peut ainsi changer la __feuille__ utilisee par defaut par les objets crees.
    # Pour changer la feuille par defaut, il suffit donc de modifier l'attribut de classe.
    # Si la __feuille__ est nulle (None), les objets ne sont pas enregistres.
    # Cela sert essentiellement a creer des objets temporaires, ou a faire de la geometrie sans figures (sic!)
    
    def __init__(self, **kw):
        self.synchroniser_feuille()

        # Parametres par defauts
        # (Note: verifier si ils sont definis, lors d'heritages multiples, il ne faut pas tout reinitialiser)
        # (exemple, lors de la creation d'un vecteur, on appelle Segment.__init__(), et Vecteur_generique.__init__())

        if not hasattr(self, "_style"):
            self._style = param.defaut_objets.copy()
                    
        if not hasattr(self, "pointable"):
            self.pointable = hasattr(self, "_coordonnees")

        if not hasattr(self, "modifiable"):
            if self.pointable and len(inspect.getargspec(self._coordonnees)[0]) > 2:
                self.modifiable = True # si _coordonnees accepte des arguments autres que 'self', l'objet est modifiable
            else:
                self.modifiable = False
        
        # Parametres identiques pour tous les objets

        self.__representation__ = []
        self.recalculer = True
        

        
# Fonctions d'initialisation de l'objet :
#########################################

    def synchroniser_feuille(self):
        if self.__feuille__:
            if self.__feuille__.parent:
                self.__class__.__feuille__ = self.__feuille__.parent.parent.actuel().actuelle
        self.__feuille__ = self.__class__.__feuille__ 
        # quand on change d'onglet dans WxGeometrie, ou de feuille courante, la feuille par defaut se synchronise (ouf!)

# La fonction suivante repertorie certaines infos utiles ( dependances...)
# ATTENTION : elle doit etre appelee A LA FIN de la fonction __init__(), pour CHAQUE objet !
# (sauf eventuellement pour les objets generiques, qui ne sont jamais crees tels quels...)
# Pour l'instant, les parametres du type **keywords ne sont PAS supportes.
# A terme, il devraient servir pour indiquer des parametres optionnels de style (couleur, ...).

    def enregistre(self, *types): # destine a remplacer le enregistre actuel
        precedent = sys._getframe(1) # contexte depuis lequel enregistre est appele.
        avant_precedent = sys._getframe(2) # contexte depuis lequel le contexte precedent est appele.
        test = avant_precedent.f_code.co_name == "__init__" and inspect.getargvalues(avant_precedent)[3].get('self') is inspect.getargvalues(precedent)[3].get('self')
        # si le test est positif, enregistre() est lance via une commande style:
        # objet.__init__(self)
        # dans ce cas, il ne faut pas donner pour l'instant de valeur a self.attributs
        # en effet, un autre enregistre suivra.


        # Effectue "self.parametre = parametre" pour tous les parametres, sauf "self":
        vars = inspect.getargvalues(precedent)[-1]
        vars.pop("self")
        kw = vars.pop("kw", {})
        self._style.update(kw)
        if param.debug:
            for key, val in vars.items():
                if hasattr(self, key) and getattr(self, key) is not val:
                    print "Attention, classe %s : le nom d'attribut '%s' est deja utilise, pour un autre objet, par une classe parente !" %(self.classe(), key)
                    print "Renommer le parametre '%s' de la methode %s.__init__()." %(key, self.classe())
        self.__dict__.update(vars)


        # on liste les parametres dans self.attributs (sert entre autres pour la sauvegarde)
        if not test:  # initialisation de l'objet
            args = inspect.getargspec(self.__init__) # arguments de la fonction __init__          
            self.attributs = args[0][1:]
            if args[1]:                 # arguments rentres sous la forme "*args"
                self.attributs.append(args[1])
                    
            dependances = self.ancetres()
            # plus self.hierarchie est faible, plus l'objet est haut place dans la hierarchie
            if dependances: self.hierarchie = max([objet.hierarchie for objet in dependances]) + 1
            else:           self.hierarchie = 0 # au sommet, les objets libres (ou suzerains)
            
            for objet in dependances:   # l'objet est vassal de chacun des objets dont il depend
                objet.vassaux.append(self)
                
            self.vassaux = WeakList()   # lors de sa creation, l'objet n'a, lui, aucun vassal (aucun objet ne depend de lui)

        
            # on verifie le type des parametres:
            for nom, type in zip(self.attributs, types): # ne le faire que lors de l'initialisation de l'objet
                obj = getattr(self, nom) # sinon, self.attributs ne correspond pas aux parametres entres en cas d'heritage
                if type == None:
                    pass
                elif isinstance(type, list):
                    for o in obj:
                        self.verifier_type(o, type[0])
                else:
                    self.verifier_type(obj, type)


    
    def verifier_type(self, obj, typ):
        if not isinstance(obj, typ):
            if hasattr(typ, "__name__"):
                s = str(typ.__name__)
            else:
                s = str(typ)
            if hasattr(obj, "__class__"):
                t = str(obj.__class__.__name__)
            else:
                t = str(type(obj))
            raise TypeError, ("L'objet %s doit etre du type %s, et non du type %s." %(str(obj), s, t))
    

# Styles et informations sur l'objet
####################################    
        
    def label(self, label = None):
        if label <> None:
            self.modifier(label = label, legende = param.LABEL) # si on change le label, celui-ci s'affiche par defaut en meme temps.
        legende = self.style("legende")
        if legende == param.NOM:
            nom = self.nom()
            if re.match("[A-Za-z]_[0-9]+$", nom):
                return "$" + nom[0] +"_{" + nom[2:] +"}$"
            else:
                return nom
        elif legende == param.LABEL:
            return self.style("label")
        else:
            return ""
    
    def style(self, nom = None, refresh = False, **kwargs):
        """- Si nom == None:
        Change le style de l'objet (mais ne rafraichit pas l'affichage par defaut).
        - Si nom est chaine de caractere:
        Recherche la propriete correspondante dans self._style."""
        if kwargs:
            self._style.update(kwargs)
            if refresh:
                self.refresh()
        if nom <> None:
            return self._style.get(nom)
        return self._style
        
    def modifier(self, **kwargs):
        "Change le style et rafraichit l'affichage."
        self.style(refresh = True, **kwargs)    
    
    
    def cacher(self):
        self.modifier(visible = False)
        self.info("cache")
        
    def voir(self):
        self.modifier(visible = True)
        self.info("visible")
    
    def fixe(self):
        self.style(fixe = True)
    
    def libre(self):
        self.style(fixe = False)
    
    def nom(self):
        if self.__feuille__: return self.__feuille__.nom_objet(self)
        return ""

        
    def type(cls):
        return cls.__name__.lower()
    
    type = classmethod(type)


    def classe(cls):
        return cls.__name__

    classe = classmethod(classe)
    
    
    def titre(cls, article = "un", point_final = True):
        """Affichage formate du nom de l'objet... (un peu de grammaire!).
        Article peut etre 'un', 'le', ou 'du', l'accord se faisant automatiquement.
        Le formatage est respecte (essayez 'un', 'UN', 'Un')."""
        
        titre = cls.__doc__.split("\n")[0].rstrip(".").lower()
        if point_final: titre += "."
        
        if not (article.lower() == "un"):
            titre = " ".join(titre.split()[1:])
            
        if article.lower() == "le":
            if titre[0] in "aeiouy":
                titre = "l'" + titre
            elif cls.genre() == "m":
                titre = "le " + titre
            else:
                titre = "la " + titre
                
        if article.lower() == "du":
            if titre[0] in "aeiouy":
                titre = "de l'" + titre
            elif cls.genre() == "m":
                titre = "du " + titre
            else:
                titre = "de la " + titre            
        
        if article.istitle():   titre = titre.capitalize()
        if article.isupper():   titre = titre.upper()
        return titre

    titre = classmethod(titre)


    def genre(cls):
        det = cls.__doc__.split()[0]
        if det.lower() == "un":
            return "m"
        else:
            return "f"

    genre = classmethod(genre)

        
    def nom_complet(self):
        return self.__class__.__name__.split("_")[0] + " " + self.nom()
        
    def ancetres(self): # retourne l'ensemble des objets dont depend l'objet.
        attributs = [getattr(self, nom) for nom in self.attributs if not isinstance(getattr(self, nom), (list, tuple))]
        listes = [getattr(self, nom) for nom in self.attributs if isinstance(getattr(self, nom), (list, tuple))]
        for liste in listes: attributs += list(liste)
        return no_twin([objet for objet in attributs if isinstance(objet, Objet)])
        
    def heritiers(self):    # retourne l'ensemble des objets qui dependent de l'objet.
        heritiers = self.vassaux.values()
        vassaux = tuple(heritiers)
        for vassal in vassaux:
            heritiers += vassal.heritiers()
        return no_twin(heritiers)

# Acces a l'environnement exterieur
###################################

    def __canvas__(self):
        return self.__feuille__ and self.__feuille__.canvas
        # __canvas__() est une methode, afin de pouvoir s'adapter a un changement de feuille
        # je sais pas si c'est encore bien utile avec les changements d'implementation...

    def message(self, message):
        if self.__feuille__:
            self.__feuille__.message(message)
        else:
            print message

    def erreur(self, message):
        self.__feuille__.erreur(message)

    def info(self, info):
        self.message(self.nom_complet() + " " + info + ".")
    
# API graphique (affichage et gestion des coordonnees)
######################################################
    
    
    # Meme pour les objets invisibles (variables), cette fonction est definie.
    # L'idee est de creer un rudiment d'API.
    def _affiche(self):
        pass

    def affiche(self):
        # si l'objet n'a pas de nom (''), c'est que c'est une construction intermediaire
        # il ne faut donc pas l'afficher.
        nom = self.nom()
        if nom and self.__canvas__():
            if self.existe() and not nom.startswith("_") and self.style("visible"):
                self._affiche() # les objets commencant par "_" ne sont pas affiches (pure convention)
                extra = self.style("extra")
                if extra:
                    for representation in self.__representation__:
                        representation.set(**extra)
            else:
                self.__canvas__().graph.supprimer(self.__representation__)
                self.__representation__ = []
        

    # Certains objets n'existent pas toujours.
    # Par exemple, l'intersection de 2 droites est parfois vide.
    # Un objet construit a partir du point d'intersection de 2 droites peut donc ne plus exister temporairement.
    # En particulier, il est important de verifier que l'objet existe avant de l'afficher.
    def existe(self):
        return False not in [obj.existe() for obj in self.ancetres()] and False not in self.conditions_existence()
        # /!\ self.conditions_existence() ne doit PAS etre evalue si un ancetre n'existe pas (resultat imprevisible)
    


    def conditions_existence(self): # conditions specifiques pour que l'objet existe
        if self.recalculer:
            self._conditions = self._conditions_existence()
        return self._conditions

    def _conditions_existence(self): # conditions specifiques pour que l'objet existe, a definir pour chaque objet
        return [] # (exemple: la mediatrice de [AB] existe ssi A <> B)


    def _espace_vital(self):
        return None
    
    def espace_vital(self):
        if self.existe():	return self._espace_vital()
    
    def equation(self, *args, **kw):
        if self.recalculer or args or kw:
            if self.existe():
                self._eq = self._equation(*args, **kw)
            else:
                self._eq = None
            self.recalculer = False
        return self._eq
        
    def coordonnees(self, *args, **kw):
        if self.recalculer or args or kw:
            if self.existe():
                self._xy = self._coordonnees(*args, **kw)
            else:
                self._xy = None # si l'objet n'existe pas, renvoit None
            self.recalculer = False
        return self._xy
    
    def val(self, *args, **kw):
        if self.recalculer or args or kw:
            if self.existe():
                self._valeur = self._val(*args, **kw)
            else:
                self._valeur = None # si l'objet n'existe pas, renvoit None
            self.recalculer = False
        return self._valeur
    
    # Notes concernant le code prcdent :
    # - Un objet qui "n'existe pas" (au sens gomtrique) a des coordonnes fixes  None.
    # - Un objet libre (c'est--dire qui accepte des arguments pour ses mthodes coordonnees() ou equation())
    #   ne doit pas avoir de conditions d'existence.
    
    def __iter__(self): # definit tuple(objet) si l'objet renvoit des coordonnees
        if hasattr(self, "_coordonnees"):
            coo = self.coordonnees()
            if coo is not None:
                return iter(coo)
        raise TypeError, "Conversion impossible, l'objet n'a pas de coordonnees."
    
    def coo(self):
        coo = array(self.coordonnees())
        if coo == None:
            self.message("L'objet %s n'est pas defini" % self.nom())
        return coo

    def position(self):
        return array(self.__canvas__().XYcoo2pix(self.coo(), -1))

    
    def refresh(self, souffler = False):
        self.recalculer = True
        heritiers = self.heritiers()
        def cmp(x, y): # relation d'ordre utilise pour le tri
            if x.hierarchie < y.hierarchie: return -1
            if x.hierarchie > y.hierarchie: return 1
            return 0
            
        heritiers.sort(cmp) # on doit recalculer les objets dans le bon ordre (suivant la hierarchie)
        for objet in heritiers:
            objet.recalculer = True  # on recalcule (au besoin) la position des objets avant d'actualiser l'affichage
        self.affiche()
        for objet in heritiers:
            objet.affiche()
        if self.__canvas__():
            self.__canvas__().actualiser()
        # dans une boucle, permet de rendre la main si necessaire au programme principal, qui reste reactif:
        if souffler:    
            wx.YieldIfNeeded()
    
    def distance_inf(self, x, y, d):
        "Teste si la distance (a l'ecran, en pixels) entre l'objet et (x, y) est inferieure a d."
        if hasattr(self, "_distance_inf") and self.existe():
            return self._distance_inf(x, y, d)
        else:
            return False    
        
    def en_gras(self):
        "Met l'objet 'en gras' (utilise quand on le selectionne)."
        e = self.style("epaisseur")
        self.style(epaisseur = 2*e)
        self.affiche()
        self.__canvas__().actualiser()
        self.style(epaisseur = e)
        self.affiche()
        
    
# API non graphique
###################
                              
    
    def _call(self, *args, **kw):
        return self.coordonnees(*args, **kw)
        
    def __call__(self, *args, **kw):
        if not self.style("fixe"):
            return self._call(*args, **kw)
        else:
            return self._call()


    def renomme(self, nom, **kw):
        "Permet de renommer l'objet, et eventuellement de changer en meme temps son style."
        nom_actuel = self.nom()
        if nom_actuel <> nom:
            self.__feuille__.objets.__verifier_nom__(nom)
            objet = self.__feuille__.objets.__dict__.pop(nom_actuel)
            self.__feuille__.objets.__dict__[nom] = objet
        self.style(**kw)
        if self.__canvas__():
            self.affiche()
            self.__canvas__().actualiser()

    def supprimer(self):
        nom = self.nom_complet() # le nom doit etre recupere AVANT la suppression
        self._supprime()
        if self.__canvas__():
            self.__canvas__().actualiser()
        self.message("%s supprime." %nom)

    
    def _supprime(self):
        for obj in self.vassaux.values():
            try: obj._supprime()
            except KeyError: pass # il se peut que l'objet n'existe deja plus.
        if self.__feuille__:
            if self.__canvas__():
                self.__canvas__().graph.supprimer(self.__representation__)
            for key in self.__feuille__.nom_objet(self, liste = True):
                self.__feuille__.objets.__dict__.pop(key)


    def __repr__(self):  # ceci sert pour sauvegarder les objets
        def formater(objet):
            if isinstance(objet, Objet):
                if objet.nom():    return objet.nom()
                #if isinstance(objet, Variable): return repr(objet.val())
            if isinstance(objet, (list, tuple)):
                return "[" + ",".join([formater(item) for item in objet]) + "]"
            return repr(objet)
                
        return self.classe() + "(" + ",".join([formater(getattr(self, attribut)) for attribut in self.attributs]) + ", **" + repr(self.style()) + ")"

        
    def copy(self):
        return self.__class__(*[getattr(self, attribut) for attribut in self.attributs])
        
    


    
# Utilitaires de conversion (permettant d'accepter plusieurs types d'entrees pour plus de souplesse).
###########################

    def nb2var(self, val):    # permet a l'utilisateur de rentrer au choix une Variable, ou un entier, ou... sans difference de traitement  (un nombre est "converti" en variable temporaire).
        if isinstance(val, Variable):
            return val.copy()
        return Variable(val)
    
    def lign2bip(self, ligne): # convertit en bipoint (couple de points) n'importe quel type de ligne.
        if isinstance(ligne, Ligne): ligne = (ligne.point1, ligne.point2)
        return ligne










################################################################################
    
# Points
    
class Point_generique(Objet):
    """Un point generique.
    
    Usage interne : la classe mere pour tous les types de points (libres, barycentres, intersection...)."""
    

    def _affiche(self):
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().Point(), self.__canvas__().text()]
        x, y = self.coordonnees()
        plot, text = self.__representation__
        plot._x = array([x]) ; plot._y = array([y])
        plot._markeredgecolor = plot._markerfacecolor= self.style("couleur")
        plot._marker = self.style("style")
        plot._markersize = self.style("taille")
        plot._markeredgewidth = self.style("epaisseur")
        plot.zorder = 3
        text._x = x + 3.*self.__canvas__().coeff(0) ; text._y = y + 2.*self.__canvas__().coeff(1)
        text._text = self.label()
        text.zorder = 4
    
    
    def _espace_vital(self):
        x, y = self.coordonnees()
        return (x, x, y, y)

    def _coordonnees(self):
        return (0.,0.)    # sera implemente differemment pour chaque type de point
        
    def abscisse(self):
        return self.coordonnees()[0]
        
    def ordonnee(self):
        return self.coordonnees()[1]
    
    def __sub__(self, y):
        return self + (-y)
    
        
    def __add__(self, y):
        if isinstance(y, Vecteur_generique): return Point_final(self, [y])
        raise TypeError, "vecteur attendu"
    
    def __gt__(self, point2): # A>B est un alias de Vecteur(A,B) - attention aux parentheses pour 2*(A>B) !
        return Vecteur(self, point2)
    
    #def _call(self, *args):
    #    return self._coordonnees(*args)
    
    def __init__(self, **kw):
        Objet.__init__(self)
        self.style(**param.points)
        
    def _distance_inf(self, x, y, d):
        x0, y0 = self.position()
        return (x - x0)**2 + (y - y0)**2 < d**2
        
    
    
    
    
class Point(Point_generique):
    """Un point libre."""

    def _coordonnees(self, x = None, y = None, auto_refresh = True):
        if isinstance(x, (tuple, list)):
            x, y = x[0], x[1]
        if x == None and y == None:
            auto_refresh = False
        x = self.x(x, False); y = self.y(y, False)
        if auto_refresh:
            self.refresh()
        return x, y        
        # False : ne rafraichit l'affichage qu'a la fin (deja que c'est lent !!).

    def __init__(self, x = None, y = None, **kw):
        if x == None:
            self.synchroniser_feuille()
            if self.__canvas__():
                xmin, xmax, ymin, ymax = self.__canvas__().fen()
                x = random.uniform(xmin,xmax)
                y = random.uniform(ymin,ymax)
                del xmin, xmax, ymin, ymax # evite d'enregistrer les variables xmin, xmax...
            else:
                x = y = 0
        
        x, y = self.nb2var(x), self.nb2var(y)
        # on convertit l'entree en Variable (voir la definition de la Classe Variable en fin de page)
        # l'entree peut-etre un nombre, un(e longueur de) segment, un (rayon de) cercle
        Point_generique.__init__(self)
        self.style(**param.points_deplacables)
        self.enregistre(Variable, Variable)



class Barycentre(Point_generique):
    """Un barycentre.
    
    Un barycentre de n points."""
    
    def _coordonnees(self):  # renvoie les coordonnees du barycentre en fonction de celles des points        
        coeffs = self.coefficients() # valeur des coefficients
        return builtin.sum(k*M.coo() for k, M in zip(coeffs, self.points))/self.__total
        
    def coefficients(self, *coeffs, **kw):
        if coeffs:
            self.verifier_coeffs(coeffs)
            for coeff, valeur in zip(self.coeffs(), coeffs):
                coeff(valeur, False) # l'objet coeff (de type "Variable") recoit la valeur "valeur".
            if kw.get("auto_refresh", True):  self.refresh()
        return [coeff() for coeff in self.coeffs]
        
    #def somme(self): return sum(self.coeffs_val())
    
    def verifier_coeffs(self, coeffs = []):
        if len(self.points) <> len(coeffs):
            self.erreur("nombre de coefficients incorrect.")                    
        if not sum([coeff() for coeff in coeffs]):
            self.erreur("la somme des coefficients est nulle.") 
    
    def _conditions_existence(self):
        self.__total = sum(self.coefficients()) # on le stocke pour eviter de le calculer 2 fois.
        return [self.__total <> 0]
    
    
    #def changer_coeffs(self, coeffs = []):  # A ADAPTER
    #    if len(self.points) <> len(coeffs):
    #        self.erreur("nombre de coefficients incorrect.")                    
    #    if not sum(coeffs):
    #        self.erreur("la somme des coefficients est nulle.") 
    #    self.coeffs = map(self.nb2var,coeffs)   
    
    def __init__(self, points, coeffs = None, **kw):
        if coeffs == None:
            coeffs = [1 for point in points]
        coeffs = map(self.nb2var,coeffs)
        Point_generique.__init__(self)        
        self.enregistre([Point_generique], [Variable])
        self.verifier_coeffs(coeffs)
        

    
class Milieu(Barycentre):
    """Un milieu.
    
    Le milieu de 2 points"""
    
    def __init__(self, point1, point2, **kw):
        Barycentre.__init__(self, points = [point1, point2])
        self.enregistre(Point_generique, Point_generique)




class Point_final(Point_generique):
    """Un point defini par une relation vectorielle.
    
    Point defini par une relation vectorielle.
                              ->      ->     ->
    Exemple : point N tel que AN = -3 AB + 2 BC
    N = Point_final(A, [A>B, B>C], [-3,2])"""
    
    # A ADAPTER POUR ACCEPTER DES VARIABLES COMME COEFFICIENTS
    
    def _coordonnees(self):
        return self.depart.coo() + builtin.sum(coeff()*vecteur.coo() for coeff, vecteur in zip(self.coeffs, self.vecteurs))


    def __init__(self, depart, vecteurs = [], coeffs = None, **kw):
        if coeffs == None:
            coeffs = [1 for vecteur in vecteurs]
        coeffs = map(self.nb2var,coeffs)
        Point_generique.__init__(self)
        self.enregistre(Point_generique, [Vecteur_generique], [Variable])
    
    
    
    
class Point_rotation(Point_generique):
    """Une image d'un point par rotation.
    
    Point construit a partir d'un autre via une rotation d'angle et de centre donne."""
    
    def _coordonnees(self):
        x0, y0 = self.rotation.centre()
        xA, yA = self.point.coordonnees()
        sina, cosa = sin(self.rotation.radian()), cos(self.rotation.radian())
        return (-sina*(yA - y0) + x0 + cosa*(xA - x0), y0 + cosa*(yA - y0) + sina*(xA - x0))

    def __init__(self, point, rotation, **kw):
        Point_generique.__init__(self)
        self.enregistre(Point_generique, Rotation)



class Projete_generique(Point_generique):
    """Un projete generique.
    
    Classe mere des differents types de projetes orthogonaux (sur droite, sur cercle ou sur segment)."""
    
    def __init__(self, point, objet, **kw):
        Point_generique.__init__(self)
        self.enregistre(Point_generique, Objet)


    def distance(self):
        return Ligne(self, self.point)._longueur()




class Projete_droite(Projete_generique):
    """Un projete orthogonal sur une droite.
    
    Projete orthogonal d'un point sur une droite."""
    
    def __init__(self, point, droite, **kw):
        Projete_generique.__init__(self, point = point, objet = droite)
        self.enregistre(Point_generique, Droite)
        
    def _conditions_existence(self):
        self.__perpendiculaire = Perpendiculaire(self.droite, self.point)
        return [self.__perpendiculaire.existe()]
    
    def _coordonnees(self):
        # il faudra prevoir le cas ou la perpendiculaire n'existe pas (si le point appartient a la droite)
        return Intersection_droites(self.droite, self.__perpendiculaire).coordonnees()
        
        
        
class Projete_cercle(Projete_generique):
    """Un projete orthogonal sur un cercle.
    
    Projete orthogonal d'un point sur un cercle."""
    
    def __init__(self, point, cercle, **kw):
        Projete_generique.__init__(self, point = point, objet = cercle)
        self.enregistre(Point_generique, Cercle_generique)

    def _conditions_existence(self):
    # cas de non existence : si le point est au centre du cercle
        self.__droite = Droite(self.point, self.cercle.centre)
        return [self.__droite.existe()]
    
    
    def _coordonnees():
        return Intersection_droite_cercle(self.__droite, self.cercle, self.point).coordonnees()
        

        
class Projete_arc_cercle(Projete_generique):
    """Un projete orthogonal sur un arc de cercle.
    
    Projete orthogonal d'un point sur un arc de cercle."""
    
    # il faudrait prevoir le cas de non existence (si le point est au centre du cercle)
    def __init__(self, point, arc, **kw):
        Projete_generique.__init__(self, point = point, objet = arc)
        self.enregistre(Point_generique, Arc_cercle)

    def _conditions_existence(self):
        self.__inter = Intersection_droite_cercle(Droite(self.point, self.arc.centre), self.arc, self.point)
        return [self.__inter.existe()]
    
    
    def _coordonnees():
        return __inter.coordonnees()



class Projete_segment(Projete_generique):
    """Un projete orthogonal sur un segment.
    
    Projete orthogonal d'un point sur un segment."""

    # il faudrait prevoir le cas de non existence (si le point est au centre du cercle)
    def __init__(self, point, segment, **kw):
        Projete_generique.__init__(self, point = point, objet = segment)
        self.enregistre(Point_generique, Segment)

    
    
    def _conditions_existence(self):
        self.__M = Intersection_droites(Droite(self.segment.point1, self.segment.point2), Perpendiculaire(self.segment, self.point))
        return [self.__M.existe()]


    def _coordonnees(self):
        A = self.segment.point1; B = self.segment.point2
        M = self.__M
        p = produit_scalaire(Vecteur(A, B), Vecteur(A, M))
        if p < 0:
            return A.coordonnees()
        if p > Vecteur(A, B)._longueur()**2:
            return B.coordonnees()
        return M.coordonnees()
        

class Projete_demidroite(Projete_generique):
    """Un projete orthogonal sur une demi-droite.
    
    Projete orthogonal d'un point sur une demi-droite."""

    # il faudrait prevoir le cas de non existence (si le point est au centre du cercle)
    def __init__(self, point, demidroite, **kw):
        Projete_generique.__init__(self, point = point, objet = demidroite)
        self.enregistre(Point_generique, Demidroite)

    
    def _conditions_existence(self):
        self.__M = Intersection_droites(Droite(self.demidroite.point1, self.demidroite.point2), Perpendiculaire(self.demidroite, self.point))
        return [self.__M.existe()]


    def _coordonnees(self):
        A = self.demidroite.point1; B = self.demidroite.point2
        M = self.__M
        p = produit_scalaire(Vecteur(A, B), Vecteur(A, M))
        if p < 0:
            return A.coordonnees()
        return M.coordonnees()


class Centre_polygone_generique(Point_generique):
    """Un centre d'un triangle.
    
    Classe mere des differents centres d'un triangle."""


    def __init__(self, polygone, **kw):
        Point_generique.__init__(self)
        self.enregistre(Polygone)



class Centre_gravite(Centre_polygone_generique):
    """un centre de gravite.
    
    Centre de gravite d'un triangle (intersection des medianes)."""
    
    def __init__(self, polygone, **kw):
        Centre_polygone_generique.__init__(self, polygone)
        self.enregistre(Polygone)

    def _conditions_existence(self):
        return [self.polygone.centre_gravite.existe()]

    def _coordonnees(self):
        return self.polygone.centre_gravite()



class Orthocentre(Centre_polygone_generique):
    """Un orthocentre.
    
    Orthocentre d'un triangle (intersection des hauteurs)."""
    
    def __init__(self, triangle, **kw):
        Centre_polygone_generique.__init__(self, triangle)
        self.enregistre(Triangle)

    def _conditions_existence(self):
        return [self.polygone.orthocentre.existe()]

    def _coordonnees(self):
        return self.triangle.orthocentre()


class Centre_cercle_circonscrit(Centre_polygone_generique):
    """Un centre du cercle circonscrit.
    
    Centre du cercle circonscrit d'un triangle (intersection des mediatrices)."""
    
    def __init__(self, triangle, **kw):
        Centre_polygone_generique.__init__(self, triangle)
        self.enregistre(Triangle)

    def _conditions_existence(self):
        return [self.polygone.centre_cercle_circonscrit.existe()]

    def _coordonnees(self):
        return self.triangle.centre_cercle_circonscrit()


class Centre_cercle_inscrit(Centre_polygone_generique):
    """Un centre du cercle inscrit.
    
    Centre du cercle inscrit d'un triangle (intersection des bissectrices)."""
    
    def __init__(self, triangle, **kw):
        Centre_polygone_generique.__init__(self, triangle)

    def _conditions_existence(self):
        return [self.polygone.centre_cercle_inscrit.existe()]

    def _coordonnees(self):
        return self.triangle.centre_cercle_inscrit()


class Centre(Point_generique):
    """Un centre d'un cercle."""
    
    def __init__(self, cercle, **kw):
        Point_generique.__init__(self)
        self.enregistre(Cercle_generique)
        
    def _conditions_existence(self):
        return [self.cercle.centre.existe()]
    
    def _coordonnees(self):
        return self.cercle.centre()
    
    
    
##########################################################################################

# Lignes



class Ligne(Objet):
    """Une ligne generique.
    
    Usage interne : la classe mere pour les droites, segments, vecteurs"""

    def __init__(self, point1, point2, **kw):
        Objet.__init__(self)
        self.enregistre(Point_generique, Point_generique)


    def _longueur(self, i = 2):
        return norm(self.point2.coo() - self.point1.coo(), i)

    
    def _espace_vital(self):
        x1, y1 = self.point1.coordonnees()
        x2, y2 = self.point2.coordonnees()
        return (min(x1, x2), max(x1,x2), min(y1, y2), max(y1,y2))
        

    def _equation(self):
        # retourne un triplet (a, b, c), tel que ax + by + c = 0 soit une equation de droite de la ligne.
        xA, yA = self.point1.coo()
        xB, yB = self.point2.coo()
        return (yA - yB, xB - xA, xA*yB - yA*xB)



class Segment(Ligne):
    """Un segment.
    
    Un segment defini par deux points"""
        
    def __init__(self, point1, point2, **kw):
        Ligne.__init__(self, point1 = point1, point2 = point2)
        self.style(**param.segments)
        self.enregistre(Point_generique, Point_generique)


    def _affiche(self):
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().plot()]
        x1, y1 = self.point1.coordonnees()
        x2, y2 = self.point2.coordonnees()
        plot = self.__representation__[0]
        plot._x = array([x1, x2]) ; plot._y = array([y1, y2])
        plot._color = self.style("couleur")
        plot._linestyle = self.style("style")
        plot._linewidth = self.style("epaisseur")

    def _distance_inf(self, x, y, d):
        # cf. "distance_point_segment.odt" dans "doc/developpeurs/maths/"
        xA, yA = self.point1.position()
        xB, yB = self.point2.position()
        x1 = min(xA, xB) - d; x2 = max(xA, xB) + d
        y1 = min(yA, yB) - d; y2 = max(yA, yB) + d
        if x1<x<x2 and y1<y<y2:
            return ((yA-yB)*(x-xA)+(xB-xA)*(y-yA))**2/((xB-xA)**2+(yB-yA)**2) < d**2
        else:
            return False
        
        
        
        A = Point(x0, y0) ; B = Point(x1, y1) ; M = Point(x, y)
        AB = Segment(A, B)
        P = Projete_segment(M, AB)
        if P.existe():
            return P.distance() < d
        else:
            return Segment(A, M)._longueur() < d



    def __contains__(self, M):
        if not isinstance(M, Point_generique):
            return False
        A = self.point1
        B = self.point2
        u = Vecteur(A, B)
        v = Vecteur(A, M)
        xu, yu = u.coordonnees()
        xv, yv = v.coordonnees()
        if abs(xu*yv-xv*yu) > param.tolerance:
            return False
        if xu:
            k = xv/xu
        elif yu:
            k = yv/yu
        else:  # A == B
            return M == A
        return 0 <= k <= 1
        
    def longueur(self):
        """Alias de _longueur, disponible pour indiquer que l'objet a vraiment
        une longueur au sens mathematique du terme (pas comme une droite !)"""
        return self._longueur()
        



class Demidroite(Ligne):
    """Une demi-droite.
    
    Une demi-droite definie par son origine et un deuxieme point"""


    def __init__(self, origine, point, **kw):
        Ligne.__init__(self, point1 = origine, point2 = point)
        self.style(**param.droites)        
        self.enregistre(Point_generique, Point_generique)
      

    def _conditions_existence(self):
        return [self.point1.coo() is not None and self.point2.coo() is not None and sum(abs(self.point1.coo() - self.point2.coo())) > param.tolerance]
        # si les conditions d'existence en amont sont mal definies, il se peut que les coordonnees d'un point valent None


    def x_y(self, nbr = 0, flag = 1):
        # si flag = 1, donne le point (x,y) sur la droite en fonction de x.
        # si flag = 0, donne le point (x,y) sur la droite en fonction de y.
        [x1, y1], [x2, y2] = self.point1.coordonnees(), self.point2.coordonnees()
        
        # si la pente est presque verticale et que flag = 1, ou presque horizontale et que flag = 0, le calcul peut exploser... l'option flag = 2 adapte flag a la pente.
        if flag == 2:  flag = self.pente(1)
        
        # dans le cas d'une droite horizontale ou verticale, flag est ignore et prend la seule valeur correcte :
        flag = not (y1 - y2) and 1 or ((x1 - x2) and flag or 0)      
        
        if flag:  return (nbr,((y1-y2)*(nbr-x1)+(x1-x2)*y1)/(x1-x2))
        else:  return (((x1-x2)*(nbr-y1)+(y1-y2)*x1)/(y1-y2),nbr)

            
    def pente(self, flag = 0):
         # si flag = 1, retourne la "tendance" de la pente (verticale : 0, horizontale : 1)
        x1, y1 = self.point1.coordonnees()
        x2, y2 = self.point2.coordonnees()
        if flag: return (abs(x2 - x1) > abs(y2 - y1)) and 1 or 0
        try: return (y2 - y1)/(x2 - x1)
        except ZeroDivisionError: return None
        
        
    def _affiche(self):
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().plot()]
        
        p = self.pente(1) # "tendance" de la pente (verticale : 0, horizontale : 1)
        fen = self.__feuille__.fenetre() # il faut tracer depuis le bord de fenetre
        x, y = self.origine.coordonnees()
        x0, y0 = self.point.coordonnees()
        x1, y1 = self.x_y((fen[2], fen[0])[p], p)
        x2, y2 = self.x_y((fen[3], fen[1])[p], p)
        plot = self.__representation__[0]

        if produit_scalaire((x1 - x, y1 - y), (x0 - x, y0 - y)) > produit_scalaire((x2 - x, y2 - y), (x0 - x, y0 - y)):
            plot._x = array([x1, x]) ; plot._y = array([y1, y])
        else:
            plot._x = array([x, x2]) ; plot._y = array([y, y2])
        plot._color = self.style("couleur")
        plot._linestyle = self.style("style")
        plot._linewidth = self.style("epaisseur")

#    def _affiche(self):
#        fen = self.__feuille__.fenetre() # il faut tracer depuis le bord de fenetre
#        p = self.pente(1)
#        A, B = self.x_y((fen[2], fen[0])[p],p), self.x_y((fen[3], fen[1])[p],p)
#        self.__canvas__().Ligne(A, B, self.style("couleur"))

    def _distance_inf(self, x, y, d):
        # cf. "distance_point_segment.odt" dans "doc/developpeurs/maths/"
        xA, yA = self.point1.position()
        xB, yB = self.point2.position()
        x1 = min(xA, xB); x2 = max(xA, xB)
        y1 = min(yA, yB); y2 = max(yA, yB)
        if (xA == x1 and x>x1-d or xA == x2 and x<x2+d) and (yA == y1 and y>y1-d or yA == y2 and y<y2+d):
            return ((yA-yB)*(x-xA)+(xB-xA)*(y-yA))**2/((xB-xA)**2+(yB-yA)**2) < d**2
        else:
            return False

    """def _distance_inf(self, x, y, d): # A REECRIRE !
        x0, y0 = self.point1.position()
        x1, y1 = self.point2.position()
        A = Point(x0, y0) ; B = Point(x1, y1) ; M = Point(x, y)
        AB = Demidroite(A, B)
        return Projete_demidroite(M, AB).distance() < d"""
        
    def __contains__(self, M):
        if not isinstance(M, Point_generique):
            return False
        A = self.origine
        B = self.point
        u = Vecteur(A, B)
        v = Vecteur(A, M)
        xu, yu = u.coordonnees()
        xv, yv = v.coordonnees()
        if abs(xu*yv-xv*yu) > param.tolerance:
            return False
        if xu:
            k = xv/xu
        elif yu:
            k = yv/yu
        else:  # A == B
            return M == A
        return 0 <= k

    def mise_en_equation(self):
        "Equation sous forme lisible par l'utilisateur."
        eq = self.equation()
        if eq is None:
            return "L'objet n'est pas defini."
        a, b, c = (str(nchiffres(coeff, param.chiffres_significatifs)) for coeff in eq)
        pente = self.pente(1) # "tendance" de la pente (verticale : 0, horizontale : 1)
        if pente:
            if (self.point.ordonnee() - self.origine.ordonnee()) > param.tolerance:
                ajout = "y > " + str(nchiffres(self.origine.ordonnee(), param.chiffres_significatifs))
            elif (self.point.ordonnee() - self.origine.ordonnee()) < param.tolerance:
                ajout = "y < " + str(nchiffres(self.origine.ordonnee(), param.chiffres_significatifs))
            else:
                return "Precision insuffisante."
        else:
            if (self.point.abscisse() - self.origine.abscisse()) > param.tolerance:
                ajout = "x > " + str(nchiffres(self.origine.abscisse(), param.chiffres_significatifs))
            elif (self.point.abscisse() - self.origine.abscisse()) < param.tolerance:
                ajout = "x < " + str(nchiffres(self.origine.abscisse(), param.chiffres_significatifs))
            else:
                return "Precision insuffisante."
        
        return "%s*x + %s*y + %s = 0 et %s" %(a, b, c, ajout) # on ne garde "que" 7 chiffres significatifs




class Droite(Ligne):
    """Une droite.
    
    Une droite definie par deux points"""

    def __init__(self, point1, point2, **kw):
        Ligne.__init__(self, point1 = point1, point2 = point2)
        self.style(**param.droites)        
        self.enregistre(Point_generique, Point_generique)
      

    def _conditions_existence(self):
        return [self.point1.coo() is not None and self.point2.coo() is not None and sum(abs(self.point1.coo() - self.point2.coo())) > param.tolerance]
        # si les conditions d'existence en amont sont mal definies, il se peut que les coordonnees d'un point valent None


    def x_y(self, nbr = 0, flag = 1):
        # si flag = 1, donne le point (x,y) sur la droite en fonction de x.
        # si flag = 0, donne le point (x,y) sur la droite en fonction de y.
        [x1, y1], [x2, y2] = self.point1.coordonnees(), self.point2.coordonnees()
        
        # si la pente est presque verticale et que flag = 1, ou presque horizontale et que flag = 0, le calcul peut exploser... l'option flag = 2 adapte flag a la pente.
        if flag == 2:  flag = self.pente(1)
        
        # dans le cas d'une droite horizontale ou verticale, flag est ignore et prend la seule valeur correcte :
        flag = not (y1 - y2) and 1 or ((x1 - x2) and flag or 0)      
        
        if flag:  return (nbr,((y1-y2)*(nbr-x1)+(x1-x2)*y1)/(x1-x2))
        else:  return (((x1-x2)*(nbr-y1)+(y1-y2)*x1)/(y1-y2),nbr)

            
    def pente(self, flag = 0):
         # si flag = 1, retourne la "tendance" de la pente (verticale : 0, horizontale : 1)
        x1, y1 = self.point1.coordonnees()
        x2, y2 = self.point2.coordonnees()
        if flag: return (abs(x2 - x1) > abs(y2 - y1)) and 1 or 0
        try: return (y2 - y1)/(x2 - x1)
        except ZeroDivisionError: return None
        
        
    def _affiche(self):
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().plot()]
        x1, y1 = self.point1.coordonnees()
        x2, y2 = self.point2.coordonnees()
        p = self.pente(1) # "tendance" de la pente (verticale : 0, horizontale : 1)
        fen = self.__feuille__.fenetre() # il faut tracer depuis le bord de fenetre
        x1, y1 = self.x_y((fen[2], fen[0])[p], p)
        x2, y2 = self.x_y((fen[3], fen[1])[p], p)
        plot = self.__representation__[0]
        plot._x = array([x1, x2]) ; plot._y = array([y1, y2])
        plot._color = self.style("couleur")
        plot._linestyle = self.style("style")
        plot._linewidth = self.style("epaisseur")

#    def _affiche(self):
#        fen = self.__feuille__.fenetre() # il faut tracer depuis le bord de fenetre
#        p = self.pente(1)
#        A, B = self.x_y((fen[2], fen[0])[p],p), self.x_y((fen[3], fen[1])[p],p)
#        self.__canvas__().Ligne(A, B, self.style("couleur"))

    def _distance_inf(self, x, y, d):
        # cf. "distance_point_segment.odt" dans "doc/developpeurs/maths/"
        xA, yA = self.point1.position()
        xB, yB = self.point2.position()
        return ((yA-yB)*(x-xA)+(xB-xA)*(y-yA))**2/((xB-xA)**2+(yB-yA)**2) < d**2

        
    """def _distance_inf(self, x, y, d):
        x0, y0 = self.point1.position()
        x1, y1 = self.point2.position()
        A = Point(x0, y0) ; B = Point(x1, y1) ; M = Point(x, y)
        AB = Droite(A, B)
        return Projete_droite(M, AB).distance() < d"""
        
    def __contains__(self, M):
        if not isinstance(M, Point_generique):
            return False
        A = self.point1
        B = self.point2
        u = Vecteur(A, B)
        v = Vecteur(A, M)
        xu, yu = u.coordonnees()
        xv, yv = v.coordonnees()
        return abs(xu*yv-xv*yu) < param.tolerance

    def mise_en_equation(self):
        "Equation sous forme lisible par l'utilisateur."
        eq = self.equation()
        if eq is None:
            return "L'objet n'est pas defini."
        a, b, c = (str(nchiffres(coeff, param.chiffres_significatifs)) for coeff in eq)
        return "%s*x + %s*y + %s = 0" %(a, b, c) # on ne garde "que" 7 chiffres significatifs




class Droite_vectorielle(Droite):
    """Une droite vectorielle.
    
    Une droite definie par un vecteur directeur (sous forme de bipoint, ou de vecteur, ou ...), et un point.
    NOTE : convertir en bipoint, puis en vecteur, permet d'accepter des droites comme argument"""
    
    def __init__(self, vecteur, point, **kw):
        vecteur = self.lign2bip(vecteur) # conversion en bipoint
        vecteur = vecteur[0]>vecteur[1]  # conversion en vecteur
        Droite.__init__(self, point1 = point, point2 = Point_final(point, [vecteur]))
        self.enregistre(Vecteur_generique, Point_generique)



class Parallele(Droite_vectorielle):
    """Une parallele.
    
    Une droite definie parallele a ... (droite, bipoint, vecteur...) passant par le point ...
    C'est presque un alias de Droite_vectorielle."""
    
    def __init__(self, droite, point, **kw):
        droite = self.lign2bip(droite) # conversion en bipoint
        droite = Droite(droite[0], droite[1])  # conversion en droite
        Droite.__init__(self, point1 = point, point2 = Point_final(point, [droite.point1>droite.point2]))
        self.enregistre(Droite, Point_generique)



class Droite_rotation(Droite):
    """Une image d'une droite par rotation.
    
    Une droite obtenue par rotation d'une autre droite, ou d'un bipoint, ou..."""

    def __init__(self, droite, rotation, **kw):
        droite = self.lign2bip(droite) # conversion en bipoint
        droite = Droite(droite[0], droite[1])  # conversion en droite
        Droite.__init__(self, point1 = Point_rotation(droite.point1, rotation), point2 = Point_rotation(droite.point2, rotation))
        self.enregistre(Droite, Rotation)



class Mediatrice(Droite_rotation):
    """Une mediatrice.
    
    La mediatrice d'un segment (ou d'un bipoint, ...)"""
    
    def __init__(self, objet1, objet2 = None, **kw):
        if objet2 == None:
            objet1, objet2 = self.lign2bip(objet1) # a la place de deux objets, on peut entrer un segment
        Droite_rotation.__init__(self, droite = (objet1, objet2), rotation = Rotation(Milieu(objet1, objet2), 90., "d"))
        self.enregistre(Point_generique, Point_generique)
    


class Perpendiculaire(Droite_vectorielle):
    """Une perpendiculaire.
    
    Une droite perpendiculaire a une autre passant par un point"""
    
    def __init__(self, droite, point, **kw):
        Droite_vectorielle.__init__(self, vecteur = Droite_rotation(droite = droite, rotation = Rotation(Point(0., 0.), 90., "d")), point= point)
        self.enregistre((Droite, Segment, Vecteur_generique, Demidroite), Point_generique)



class Droite_equation(Droite):
    """Une droite definie par une equation.
    
    Une droite d'equation donnee sous forme d'un triplet (a, b, c). (ax + by + c = 0)"""
    
    def __init__(self, coeffs = None, **kw):
        Droite.__init__(self, point1 = Point(0, 0), point2 = Point(1, 0))
        self.enregistre(tuple)
        self.equation(coeffs, False)

        
    def verifier_coeffs(self, coeffs = []):
        if len(coeffs) <> 3:
            self.erreur("nombre de coefficients incorrect.")                    
        if coeffs[0] == coeffs[1] == 0:
            self.erreur("les deux premiers coefficients sont nuls.") 
    
    
    def _equation(self, coeffs = None, auto_refresh = True):
        # Retourne un triplet (a, b, c), tel que ax + by + c = 0 soit une equation de droite de la ligne.
        # On peut aussi modifier l'equation de la droite.
        if coeffs <> None:
            self.verifier_coeffs(coeffs)
            a, b, c = coeffs
            self.coeffs = coeffs
            if b:
                xA = 0; yA = -c/b
                xB = 1; yB = -a/b - c/b
            else:
                xA = xB = -c/a
                yA = 0 ; yB = 1
            self.point1.coordonnees(xA, yA, False)
            self.point2.coordonnees(xB, yB, auto_refresh)
        xA, yA = self.point1.coo()
        xB, yB = self.point2.coo()
        
        return (yA - yB, xB - xA, xA*yB - yA*xB)
    


class Bissectrice(Mediatrice):
    """Une bissectrice.
    
    La bissectrice d'un angle defini par 3 points.
    On peut, au choix, entrer un angle ou 3 points comme argument."""
    
    def __init__(self, pt1, pt2 = None, pt3 = None, **kw):
        if pt2 == None: # point1 est un angle
            pt2 = pt1.point2
            pt3 = pt1.point3
            pt1 = pt1.point1
        
        
        A = pt2 + Vecteur_unitaire(pt2>pt1)
        B = pt2 + Vecteur_unitaire(pt2>pt3)
        Mediatrice.__init__(self, A, B)
        #del A, B
        self.enregistre(Point_generique, Point_generique, Point_generique)

    def _conditions_existence(self):
        if None in (self.pt1.coo(), self.pt2.coo(), self.pt3.coo()):
            return [False]
        return [   sum(abs(self.pt1.coo() - self.pt2.coo())) > param.tolerance,
                    sum(abs(self.pt2.coo() - self.pt3.coo())) > param.tolerance,
                    sum(abs(self.pt1.coo() - self.pt3.coo())) > param.tolerance]






class Tangente(Perpendiculaire):
    """Une tangente.
    
    Une des deux tangente a un cercle passant par un point exterieur au cercle.
    Le dernier parametre (True/False) sert a distinguer les deux tangentes.
    (Voir la classe Intersection_cercles pour plus d'infos)."""
    
    def __init__(self, cercle, point, angle_positif = None, **kw):
        self.synchroniser_feuille()
        if self.__feuille__:            # si l'on cree 2 fois de suite un tangente de memes cercle et point,
            if angle_positif == None:   # alors on doit obtenir les deux tangentes differentes possibles.
                tangentes = [objet for objet in self.__feuille__.objets() if isinstance(objet, Tangente) and objet.cercle is cercle and objet.point is point]
                if tangentes:
                    angle_positif = not tangentes[0].angle_positif # on cree l'autre tangente
                else:
                    angle_positif = True
                del tangentes
        else:
            angle_positif = True
        
        # on a deux cas :
        # si le point est sur le cercle, on prend le point (Attention a bien prevoir tous les cas)
        # si le point n'est pas sur le cercle, on construit une intersection de cercles
        
        # la dichotomie n'est pas sensas, car il se peut que M soit tres pret du cercle...
        if isinstance(point, Glisseur_cercle) and point.cercle is cercle or isinstance(cercle, Cercle_diametre) and (cercle.point1 is point or cercle.point2 is point) or isinstance(cercle, Cercle_rayon) and cercle.point is point:
            M = point
        else:
            M = Intersection_cercles(cercle, Cercle_diametre(cercle.centre, point), angle_positif)
        Perpendiculaire.__init__(self, Droite(cercle.centre, M), M)
        del M
        self.enregistre(Cercle_generique, Point_generique)
        


##########################################################################################

# Vecteurs


class Vecteur_generique(Objet):
    """Un vecteur generique.
    
    Usage interne : la classe mere pour les differents vecteurs"""

    def __init__(self, **kw):
        Objet.__init__(self)

    def _coordonnees(self):
        return self.point2.coo()- self.point1.coo()
        
    def abscisse(self):
        return self.coordonnees()[0]
        
    def ordonnee(self):
        return self.coordonnees()[1]


    def norme(self, i = 2):
        return norm(self.coo(), i)

    def __add__(self, y):
        if isinstance(y, Vecteur_generique):
            return Somme_vecteurs([self, y])
        raise TypeError, "vecteur attendu"
    
    def __sub__(self, y):
        return self + (-y)
    
    def __rmul__(self, y):
        return Somme_vecteurs([self], [y])

    #def _call(self, *args):
    #    return self._coordonnees(*args)
    
    def __neg__(self, *args):
        return (-1)*self

    def __div__(self, y):
        return (1./y)*self
        
    def __truediv__(self, y):
        return self.__div__(y)
    
    def __eq__(self, y):
        return isinstance(y, Vecteur_generique) and tuple(self.coordonnees()) == tuple(y.coordonnees())
    
    def __nonzero__(self):
        return tuple(self.coordonnees()) <> (0, 0)
        
    def __ne__(self, y):
        return not self == y



class Vecteur_rotation(Point_rotation, Vecteur_generique): # en pratique, c'est presque un alias de Point_rotation...
    """Une image d'un vecteur par une rotation.
    
    Vecteur construit a partir d'un autre via une rotation d'angle et de centre donne"""
    
    def __init__(self, vecteur, rotation, **kw):
        Point_rotation.__init__(self, point = vecteur, rotation = rotation)
        self.enregistre(Vecteur_generique, Rotation)
      
    def _espace_vital(self):
        pass

    def _affiche(self):  # pour que le vecteur ne soit pas affich comme point
        pass


class Vecteur(Segment, Vecteur_generique):
    """Un vecteur.
    
    Un vecteur defini par deux points"""
    
    def _affiche(self):
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().plot(), self.__canvas__().plot()]
        plot1, plot2 = self.__representation__
        x1, y1 = self.point1.coordonnees()
        x2, y2 = self.point2.coordonnees()
        A = array(self.__canvas__().XYcoo2pix(array((x1, y1)), -1))
        B = array(self.__canvas__().XYcoo2pix(array((x2, y2)), -1))

        #                                                 M
        #                                               C +--\
        # l'ideal est ici un dessin:   A +---------------+----+ B
        #                                                 +--/
        #                                                 N
        # On calcule les coordonnees de C, puis par rotation autour de B, celles de M et N.
        
        #A, B = self.point1.position(), self.point2.position()
        r, a = self.style("taille"), pi/6.
        C = B + r*(A - B)/norm(A - B, 2)
        M = (dot([[cos(a), -sin(a)], [sin(a), cos(a)]], transpose([C - B])) + transpose([B]))[:,0]
        N = (dot([[cos(a), sin(a)], [-sin(a), cos(a)]], transpose([C - B])) + transpose([B]))[:,0]
        x, y = self.__canvas__().XYpix2coo(zip(M, B, N), -1)
        plot1._x = array([x1, x2]) ; plot1._y = array([y1, y2])
        plot2._x = array(x) ; plot2._y = array(y)
        plot1._color = plot2._color = self.style("couleur")
        plot1._linestyle = plot2._linestyle = self.style("style")
        plot1._linewidth = plot2._linewidth = self.style("epaisseur")


    def __init__(self, point1, point2, **kw):
        Segment.__init__(self, point1 = point1, point2 = point2)
        self.style(**param.vecteurs)
        Vecteur_generique.__init__(self)
        self.enregistre(Point_generique, Point_generique)
    
    
class Vecteur_unitaire(Vecteur):
    """Un vecteur unitaire.
    
    Un vecteur defini en normalisant un autre vecteur.
    Il aura donc meme sens et meme direction, mais sera de norme 1."""

    def __init__(self, vecteur, **kw):
        Vecteur_generique.__init__(self)
        self.enregistre(Vecteur_generique)
    
    def _coordonnees(self):
        return (self.vecteur.abscisse()/self.__norme, self.vecteur.ordonnee()/self.__norme)
    
    def _conditions_existence(self):
        self.__norme = self.vecteur.norme()
        return [self.__norme > param.tolerance]
        
    

class Vecteur_libre(Vecteur):
    """Un vecteur libre.
    
    Un vecteur defini par ses coordonnees."""

    def __init__(self, x = None, y = None, **kw):
        if x == None:
            self.synchroniser_feuille()
            if self.__feuille__:
                xmin, xmax, ymin, ymax = self.__canvas__().fen()
                x = random.uniform(xmin,xmax)
                y = random.uniform(ymin,ymax)
                del xmin, xmax, ymin, ymax # evite d'enregistrer les variables xmin, xmax...
            else:
                x = y = 0
        
        x, y = self.nb2var(x), self.nb2var(y)
        # on convertit l'entree en Variable (voir la definition de la Classe Variable en fin de page)
        # l'entree peut-etre un nombre, un(e longueur de) segment, un (rayon de) cercle
        Vecteur.__init__(self, point1 = Point(0., 0.), point2 = Point(x, y))
        self.enregistre(Variable, Variable)
        
    def _affiche(self):
        pass  # les vecteurs libres ne sont pas affiches.
    
    def _coordonnees(self, x = None, y = None):
        return self.point2._coordonnees(x, y)

    
    
    
    
class Somme_vecteurs(Vecteur_generique):
    """Une somme de vecteurs.
    
    Les arguments sont des couples coefficient, vecteur
    Exemple: Somme_vecteurs([u,v,A>B], [3,2,-5])."""
    
    def __init__(self, vecteurs = [], coeffs = None, **kw):
        
        if coeffs == None:
            coeffs = [1 for vecteur in vecteurs]
        coeffs = map(self.nb2var, coeffs)
        
        _coeffs_, _vecteurs_ = [], [] # le code suivant evite de multiplier les sommes de vecteurs imbriquees
        for i in range(len(vecteurs)):
            if isinstance(vecteurs[i], Somme_vecteurs):
                _coeffs_.extend(list(coeffs[i]*array(vecteurs[i].coeffs)))
                _vecteurs_.extend(vecteurs[i].vecteurs)
            else:
                _coeffs_.append(coeffs[i])
                _vecteurs_.append(vecteurs[i])
        coeffs, vecteurs = _coeffs_, _vecteurs_
        
        Vecteur_generique.__init__(self)
        self.enregistre([Vecteur_generique], [Variable])
        
    
    def _coordonnees(self):
        return sum([self.coeffs[i].val()*self.vecteurs[i].coo() for i in range(len(self.coeffs))])


         
        
        
##########################################################################################

# Cercles


class Arc_cercle(Objet):
    """Un arc de cercle.
    
    Un arc de cercle oriente, defini par son centre et ses extremites, dans le sens direct."""
    
    
    def __init__(self, centre, point1, point2, **kw):
        Objet.__init__(self)
        vecteur_unite = Vecteur_libre(1, 0)
        angle1 = Angle_oriente(vecteur_unite, centre>point1)
        angle2 = Angle_oriente(vecteur_unite, centre>point2)
        angle_au_centre = Angle_oriente(centre>point1, centre>point2)
        self.style(**param.cercles)
        self.enregistre(Point_generique, Point_generique, Point_generique)

    def rayon(self):
        return Segment(self.centre, self.point1)._longueur()

    def _affiche(self):  
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().plot()]
        
        plot = self.__representation__[0]
        x, y = self.centre()
        r = self.rayon()
        a = self.angle1.val()
        b = self.angle2.val()
        if b < a:
            b += 2*pi
        t = concatenate((arange(a, b, self.__canvas__().pas()), [b]))
        plot._x = x + r*cos(t)
        plot._y = y + r*sin(t)
        plot._color = self.style("couleur")
        plot._linestyle = self.style("style")
        plot._linewidth = self.style("epaisseur")


    def diametre(self):
        return 2*self.rayon()
    
    def longueur(self):
        a = self.angle_au_centre.val()
        if a < 0:
            a += 2*pi
        return a*self.rayon()
    
    _longueur = longueur
    
    def _equation(self):
        # retourne a, b, c tels que x**2 + y**2 + ax + by + c = 0
        a, b = self.centre(); c = self.rayon()
        return -2*a, -2*b, a**2 + b**2 - c**2

    def _distance_inf(self, x, y, d):
        x0, y0 = self.centre.position()
        a = self.angle1.val()
        b = self.angle2.val()
        if b < a:
            b += 2*pi
        t = arange(a, b, .003)
        rx, ry = [self.rayon()/self.__canvas__().coeff(i) for i in (0, 1)]
        u = x - x0 - rx*cos(t)
        v = y - y0 + ry*sin(t)
        m = min(u*u + v*v) # u*u est beaucoup plus rapide que u**2 (sic!)
        return m < d**2

    def __contains__(self, M):
        O = self.centre
        a = self.angle1.val()
        b = self.angle2.val()
        if b < a:
            b += 2*pi
        u = O>M
        c = Angle_oriente(self.vecteur_unite, u).val()
        if c < a:
            c += 2*pi
        return abs(u.norme() - self.rayon()) < param.tolerance and a < c < b


    def _espace_vital(self):
        x0, y0 = self.centre.coordonnees()
        a = self.angle1.val()
        b = self.angle2.val()
        if b < a:
            b += 2*pi
        t = arange(a, b, .003)
        r = self.rayon()
        u = x0 + r*cos(t)
        v = y0 + r*sin(t)
        return min(u), max(u), min(v), max(v)
    
    
    def mise_en_equation(self):
        "Equation sous forme lisible par l'utilisateur."
        eq = self.equation()
        if eq is None:
            return "L'objet n'est pas defini."
        a, b, c = (str(nchiffres(coeff, param.chiffres_significatifs)) for coeff in eq) # on ne garde "que" 7 chiffres significatifs
        xmin, xmax, ymin, ymax = (str(nchiffres(extremum, param.chiffres_significatifs)) for extremum in self._espace_vital())
        return "x^2 + y^2 + %s*x + %s*y + %s = 0 avec %s<x<%s et %s<y<%s" %(a, b, c, xmin, xmax, ymin, ymax)




class Demicercle(Arc_cercle):
    """Un demi-cercle.
    
    Un demi-cercle oriente, defini par ses extremites, dans le sens direct."""
    
    
    def __init__(self, point1, point2, **kw):
        Arc_cercle.__init__(self, Milieu(point1, point2), point1, point2)
        self.style(**param.cercles)
        self.enregistre(Point_generique, Point_generique, Point_generique)





class Cercle_generique(Objet):
    """Un cercle generique.
    
    Usage interne : la classe mere pour tous les types de cercles"""
    

    def _affiche(self):  
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().plot()]
        
        plot = self.__representation__[0]
        x, y = self.centre()
        r = self.rayon()
        t = concatenate((arange(0, 2*pi , self.__canvas__().pas()), [0]))
        plot._x = x + r*cos(t)
        plot._y = y + r*sin(t)
        plot._color = self.style("couleur")
        plot._linestyle = self.style("style")
        plot._linewidth = self.style("epaisseur")


#     def position(self):
#         return self.coord2pix(self.centre())
    
    #def rayon(self):    # sera implemente differemment pour chaque type de cercle
    #    return 0
    
    #def centre(self):    # sera implemente differemment pour chaque type de cercle (?)
    #    return self._centre.coordonnees()

    def perimetre(self):
        return 2*pi*self.rayon()
    
    _longueur = perimetre

    def diametre(self):
        return 2*self.rayon()
    
    def __init__(self, **kw):
        Objet.__init__(self)
        self.style(**param.cercles)
    
    def _equation(self):
        # retourne a, b, c tels que x**2 + y**2 + ax + by + c = 0
        a, b = self.centre(); c = self.rayon()
        return -2*a, -2*b, a**2 + b**2 - c**2

    def _distance_inf(self, x, y, d):
        x0, y0 = self.centre.position()
        t = arange(0, 2*pi, .003)
        rx, ry = [self.rayon()/self.__canvas__().coeff(i) for i in (0, 1)]
        u = x - x0 - rx*sin(t)
        v = y - y0 + ry*cos(t)
        m = min(u*u + v*v) # u*u est beaucoup plus rapide que u**2 (sic!)
        return m < d**2

    def __contains__(self, M):
        O = self.centre
        return abs(Segment(O, M)._longueur() - self.rayon()) < param.tolerance


    def _espace_vital(self):
        x, y = self.centre.coordonnees()
        r = self.rayon()
        return (x-r, x+r, y-r, y+r)
    
    
    def mise_en_equation(self):
        "Equation sous forme lisible par l'utilisateur."
        eq = self.equation()
        if eq is None:
            return "L'objet n'est pas defini."
        a, b, c = (str(nchiffres(coeff, param.chiffres_significatifs)) for coeff in eq)
        return "x^2 + y^2 + %s*x + %s*y + %s = 0" %(a, b, c)  # on ne garde "que" 7 chiffres significatifs





class Cercle(Cercle_generique):
    """Un cercle.
    
    Un cercle defini par son centre et son rayon"""
    
    #def rayon(self, ray = None):  # retourne le rayon (eventuellement apres l'avoir modifie)
    #    return self._rayon.val(ray)
    
    def __init__(self, centre, rayon = None, **kw):
        
        if rayon == None:
            self.synchroniser_feuille()
            if self.__feuille__:
                xmin, xmax, ymin, ymax = self.__canvas__().fen()
                rayon = random.uniform(0, min(abs(xmin-xmax), abs(ymin-ymax)))
            else:
                rayon = 1
        rayon = self.nb2var(rayon)
        Cercle_generique.__init__(self)
        self.enregistre(Point_generique, Variable)
     
     
     
class Cercle_rayon(Cercle_generique):
    """Un cercle defini par un rayon.
    
    Un cercle defini par un rayon [OM]"""
        
    def rayon(self):
        return Segment(self.centre, self.point)._longueur()
    
    def __init__(self, centre, point, **kw):
        Cercle_generique.__init__(self)
        self.enregistre(Point_generique, Point_generique)
        #self.centre = self.point1 = centre
        #self.point = self.point2 = point
        #self.rayon = Variable()
    
    
    
class Cercle_diametre(Cercle_generique):
    """Un cercle defini par un diametre.
    
    Un cercle defini par un diametre [AB]"""
    
    
    def rayon(self):
        return Segment(self.centre, self.point2)._longueur()
    
    def __init__(self, point1, point2, **kw):
        Cercle_generique.__init__(self)
        centre = Milieu(point1, point2)
        self.enregistre(Point_generique, Point_generique)
        #self.point1 = point1
        #self.point2 = point2
        #self.centre = Milieu(point1, point2)
        #self.rayon = Variable(Segment(self.centre, point2))



##########################################################################################

# Divers

# LE CODE A ETE MODIFIE DE MANIERE A INTRODUIRE SYSTEMATIQUEMENT UNE "VARIABLE" COMME ARGUMENT DES DIFFERENTS OBJETS.
# (Sauf pour les objets non rataches a une __feuille__ - objets temporaires en general)





class Variable(Objet):
    """Une variable.
    
    Une variable numerique ; l'argument peut etre un nombre, ou une expression sous forme de chaine de caracteres.
    Exemple: Variable(17.5), Variable('AB.longeur()+1').
    Dans ce dernier cas, il est necessaire qu'une feuille de travail soit definie."""
    
    def __init__(self, valeur = 0):
            
        #debug(valeur)
        #if type(valeur) ==  str: raise TypeError
        self.attributs = []
        self.vassaux = WeakList()   # lors de sa creation, l'objet n'a aucun vassal (aucun objet ne depend de lui)
        self.recalculer = True
        self.val(valeur, False)
        Objet.__init__(self)
        # l'enregistrement des variables est tres particulier, et se fait avec self.val...
    
    
        #self.enregistre((int, float, long, Cercle_generique, Segment, Glisseur_generique))
    def soustype(self):
        return isinstance(self.valeur, list) and "compose" or "simple"

    def _val(self, valeur = None, auto_refresh = True):
        """val() propose une interface standard pour afficher une valeur
        (ou la modifier, si c'est une valeur libre), indpendamment de son origine.
        
        Pour comprendre ce qui se passe, il faut savoir que le contenu de la variable est stock dans self.valeur.
        S'il s'agit d'un nombre, il est stock tel quel, mais s'il s'agit d'une expression,
        le code qui suit va remplacer toutes les noms d'objets par les objets eux-mmes.
        Ainsi, si un objet est renomm, le contenu reste pertinent.
        Exemple :
        "u.norme() + v.norme()" devient ["", u, ".norme() + ", v, ".norme()"]
        La connaissance de ce fonctionnement interne est rarement requise, mais peut-tre utile
        pour court-circuiter le fonctionnement normal de Variable par exemple.
        Voir la classe Bissectrice par exemple.
        """
                
        if valeur <> None:
            if isinstance(valeur, Variable): # si u est une variable, Variable(u) renvoit une copie de u
                val = valeur.valeur
                if isinstance(val, list):
                    val = str(valeur)
                valeur = val

            # on supprime la variable de la liste des vassaux pour les objets dont elle ne dependra plus desormais:
            for objet in self.ancetres(): 
                objet.vassaux.remove(self)
            
            self.attributs = ["valeur"]
            
            if isinstance(valeur, str):
                #debug(valeur)
                liste = regexp("[^.A-Za-z0-9_][A-Za-z_][A-Za-z0-9_]*", " " + valeur.replace(" ",""), "x[0]+'$'+x[1:]+'$'").split("$")
                #debug(liste)
                for i in range(len(liste)):
                    if i%2:
                        if liste[i]:
                            liste[i] = self.__feuille__.objets[liste[i]]
                valeur = liste

            if isinstance(valeur, (list, tuple)):
                self.valeur = valeur[:]
                dependances = self.ancetres()
                # plus self.hierarchie est faible, plus l'objet est haut place dans la hierarchie
                if dependances: self.hierarchie = max([objet.hierarchie for objet in dependances]) + 1
                else:           self.hierarchie = 0 # au sommet, les objets libres (ou suzerains)
                
                for objet in dependances:   # l'objet est vassal de chacun des objets dont il depend
                    objet.vassaux.append(self)
                    
            elif isinstance(valeur, (int, float, long)):
                self.valeur = valeur
                self.hierarchie = 0
            
            else:
                raise TypeError, "Les arguments d'une variable doivent etre de type str, int, float ou list, et non %s" %type(valeur)
            if auto_refresh:  self.refresh()

                
        if self.soustype() == "compose":
            liste = self.valeur[:]
            for i in range(len(liste)):
                if i%2: # annuaire inverse ;) (on retrouve le nom de l'objet)
                    liste[i] = [key for key, value in self.__feuille__.objets.__dict__.items() if value is liste[i]][0]
            #debug(liste)
            val = eval("".join(liste), self.__feuille__.objets.__dict__)
            if isinstance(val, Variable): # Variable("u+1") devient un alias pratique de Variable("u()+1")
                val = val.val()           # (mais d'un autre cote, ca entretient la confusion entre u et u()...)
            return val
        else:
            return 1.*self.valeur
            


    _call = _val
            
    def __repr__(self):   # donne une representation affichable de l'objet qui sert a generer la variable
        if isinstance(self.valeur, (int, float, long)):
            return "Variable(%s)" %self.val()
        else:
            liste = self.valeur[:]
            for i in range(len(liste)):
                if i%2:
                    liste[i] = [key for key, value in self.__feuille__.objets.__dict__.items() if value is liste[i]][0]
            return "Variable('%s')" % "".join(liste).strip()
            

    
    def __str__(self):
        if isinstance(self.valeur, (int, float, long)):
            return str(self.val())
        else:
            liste = self.valeur[:]
            for i in range(len(liste)):
                if i%2:
                    liste[i] = [key for key, value in self.__feuille__.objets.__dict__.items() if value is liste[i]][0]
            return "".join(liste).strip() # renvoit la chaine qui sert a initialiser la variable
    
    
    
#    def varier(self, debut = 0, fin = 1, pas = 0.02, periode = 0.03):
#        if self.__canvas__():
#            self.variation = [wx.Timer(self.__canvas__()), debut, fin, pas]
#            self.variation[0].Start(1000*periode)
#            
#            self.__canvas__().Bind(wx.EVT_TIMER, self._varier)
        
#    def _varier(self, event = None):
#        self(self.variation[1])
#        self.variation[1] += self.variation[3]
#        if self.variation[3]*(self.variation[1] - self.variation[2]) > 0:
#        self.variation[0].Stop()


# Cette portion de code marche, mais son comportement des que je modifie qqchose est surrealiste !!!        
        

    def varier(self):
        self.timer = wx.FutureCall(1, self.nvarier)
        
    def nvarier(self):
        self.bidon = 3

#    def copy(self):
#        if self.type() == "compose":
#            return Variable(repr(
    
    
        
#    def __add__(self, y):
#        if isinstance(y, (float, long, int)):
#            return Variable("%s()+%s" % (self.nom(), y))
#        if isinstance(y, self.__class__):
#            return Variable("%s()+%s()" % (self.nom(), y.nom()))
#        raise TypeError
    
#    def __radd__(self, y):
#        if isinstance(y, (float, long, int)):
#            return Variable("%s+%s()" % (y, self.nom()))
#        if isinstance(y, self.__class__):
#            return Variable("%s()+%s()" % (self.nom(), y.nom()))
#        raise TypeError
        
    
    # On definit __add__, __radd__, __sub__, etc...
#    for operation, symbole in (("add", "+"), ("sub", "-"), ("mul", "*"), ("div", "*1./"), ("pow", "**")):

#        def __op__(self, y, operation = "__" + operation + "__", symbole = symbole):
#            msg = "unsupported operand types for %s: 'Variable' and '%s'" %(symbole, type(y))
#            if self.type() == "simple":
#                if isinstance(y, (float, long, int)):
#                    return Variable(getattr(self(), operation)(y))
#                elif isinstance(y, Variable):
#                    if y.type() == "simple":
#                        return Variable(getattr(self(), operation)(y()))
#                    else:
#                        return Variable("%s%s%s" % (self, symbole, y))
#                else:
#                    raise TypeError, msg
#            elif isinstance(y, (float, long, int, Variable)):
#                return Variable("%s%s%s" % (self, symbole, y))
#            else:
#                raise TypeError, msg
#    
#        def __rop__(self, y, operation = "__r" + operation + "__", symbole = symbole):
#            msg = "unsupported operand types for %s: '%s' and 'Variable'" %(symbole, type(y))
#            if self.type() == "simple":
#                if isinstance(y, (float, long, int)):
#                    return Variable(getattr(self(), operation)(y))
#                elif isinstance(y, Variable):
#                    if y.type() == "simple":
#                        return Variable(getattr(self(), operation)(y()))
#                    else:
#                        return Variable("%s%s%s" % (y, symbole, self))
#                else:
#                    raise TypeError, msg
#            elif isinstance(y, (float, long, int, Variable)):
#                return Variable("%s%s%s" % (y, symbole, self))
#            else:
#                raise TypeError, msg
            
    for operation, symbole in (("add", "+"), ("sub", "-"), ("mul", "*"), ("div", "/"), ("truediv", "/"), ("pow", "**")):

        def __op__(self, y, operation = "__" + operation + "__"):
            return getattr(float(self), operation)(float(y))
    
        def __rop__(self, y, operation = "__r" + operation + "__"):
            return getattr(float(self), operation)(float(y))

        def __iop__(self, y, op = operation, symbole = symbole):
            self.val(getattr(float(self), "__%s__" %op)(float(y)), False)
            return self
        
        exec("__%s__=__op__; __r%s__=__rop__; __i%s__=__iop__" %(operation, operation, operation))
        
    del __op__, __rop__
        
    
#    for nom in fonctions_matplotlib:
#        exec("""
#def %s(self, *args):
#    if self.type() == "simple":
#    arguments = []
#    for arg in args:
#        arguments.append(float(arg))
#    return Variable(%s(float(self), *arguments))""" %(nom, nom))

    def __float__(self):
        return self.val()
        
    def __int__(self):
        return int(float(self))
        
    def __neg__(self):
        return 0 - self
        
    def __pos__(self):
        return 0 + self

    def __eq__(self, y):
        #print type(self), type(y)
        return hasattr(y, "__float__") and float(self) == float(y)
        
    def __ne__(self, y):
        return not self == y
        
    def __nonzero__(self):
        return self <> 0
        
    def __gt__(self, y):
        return hasattr(y, "__float__") and float(self) > float(y)
        
    def __ge__(self, y):
        return self > y or self == y




class Rotation(Objet):
    """Une rotation.
    
    Une rotation definie par son centre, et un angle en radian (r), grad (g) ou degre (d)"""
    
    def __init__(self, centre = None, valeur = pi, unite = "r", **kw):
        # l'unite par defaut est le radian.
        if centre == None:
            centre = Point(0, 0)
        valeur = self.nb2var(valeur)
        Objet.__init__(self)
        self.enregistre(Point_generique, Variable, str)
        
    #def centre(self):
    #    return self._centre.coordonnees()
    
    def angle(self, valeur = None, unite = "", auto_refresh = True):
        self.unite = unite or self.unite
        return self.valeur(valeur, auto_refresh)
        
    def radian(self):
        conversion = {"r":1., "g":pi/200., "d": pi/180.}
        return self.angle()*conversion[self.unite]


# A DEVELOPPER (image d'un objet avec __call__) !!!


class Produit_scalaire(Objet):
    """Un produit scalaire.
    
    Le produit scalaire de deux vecteurs"""
    
    def __init__(self, vecteur1, vecteur2, **kw):
        Objet.__init__(self)
        self.enregistre(Vecteur_generique, Vecteur_generique)

    def _val(self):
        return sum(self.vecteur1.coo()*self.vecteur2.coo())
    
    def __call__(self):
        return self.val()

# NOTE :
# A l'avenir, une autre methode pourrait etre d'implementer le produit scalaire comme multiplication directement dans l'objet vecteur (def __mul__(self, vecteur): etc...).
# On pourrait aussi imaginer que 2*un vecteur soit defini, ou encore vecteur + vecteur mais c'est un peu complique.
# Il y a differents types de vecteurs avec des dependances differentes, quel serait le type final ?
    
    

# A FAIRE :
# - GERER LES DIFFERENTES UNITES POUR LES ANGLES.
# - POUVOIR ENTRER UN ANGLE COMME ARGUMENT POUR LES ROTATIONS.

class Angle_generique(Objet):
    """Un angle generique.
    
    Classe mere des angles orientes ou non. L'unite par defaut est le radian."""
    
    def __init__(self, vecteur1, vecteur2, unite = "r", **kw):
        # l'unite par defaut est le radian.
        Objet.__init__(self)
        self.enregistre(Vecteur_generique, Vecteur_generique, str)
        
    def _val(self):
        angle_geomq = arccos(produit_scalaire(self.vecteur1, self.vecteur2)/(self.vecteur1.norme()*self.vecteur2.norme()))
        x1, y1 = self.vecteur1.coo(); x2, y2 = self.vecteur2.coo()
        determinant = x1*y2-x2*y1
        if determinant:
            signe = determinant/abs(determinant)
        else:
            signe = 1 # Peut importe le signe quand les vecteurs sont colineaires (0 = -0 et pi = -pi [2pi])
        conversion = {"r":1., "g":200/pi, "d": 180/pi}
        return signe*angle_geomq*conversion[self.unite]
        

    def __call__(self):
        return self.val()
    
    
    def radian(self):
        conversion = {"r":1., "g":pi/200., "d": pi/180.}
        return self.val()*conversion[self.unite]


    def degre(self):
        conversion = {"r":180/pi, "g":180/200., "d": 1}
        return self.angle()*conversion[self.unite]


    def grad(self):
        conversion = {"r":200/pi, "g":1, "d": 200/180.}
        return self.angle()*conversion[self.unite]

    def _conditions_existence(self):
        return [self.vecteur1.norme() > param.tolerance, self.vecteur2.norme() > param.tolerance]


class Angle(Angle_generique):
    """Un angle.
    
    Un angle non oriente, defini par 3 points A, B, C -> angle /ABC\."""

    def __init__(self, point1, point2, point3, unite = "r", **kw):
        # l'unite par defaut est le radian.
        Angle_generique.__init__(self, point2>point1, point2>point3)
        self.enregistre(Point_generique, Point_generique, Point_generique)

    def _val(self):
        return abs(Angle_generique.val())
        
        

class Angle_oriente(Angle_generique):
    """Un angle oriente.
    
    Un angle oriente defini par deux vecteurs."""



class Texte(Objet):
    """Un texte.
    
    Un texte a afficher"""
    
    
    
    try:
        __empty_font__ = matplotlib.font_manager.FontProperties()
    except NameError:
        pass
    
    def _coordonnees(self, x = None, y = None, auto_refresh = True):
        if isinstance(x, (tuple, list)):
            x, y = x[0], x[1]
        if x == None and y == None:
            auto_refresh = False
        x = self.x(x, False); y = self.y(y, False)
        if auto_refresh:
            self.refresh()
        return x, y        
        # False : ne rafraichit l'affichage qu'a la fin (deja que c'est lent !!).

    def _affiche(self):
        #self.__canvas__().Texte(self.texte, self.coo())
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().text(va = "center", ha = "center", zorder = 4)]
            self.__representation__[0]._fontproperties = copy.copy(self.__empty_font__)
        x, y = self.coordonnees()
        text = self.__representation__[0]
        text._x = x ; text._y = y
        text._color = self.style("couleur")
        font = text._fontproperties
        font.set_size(self.style("taille"))
        #font.set_stretch(self.style("largeur"))  # mal gere par matploltib (version 0.87)
        font.set_weight(100*min(9, max(1, int(self.style("epaisseur")))))
        font.set_style(self.style("style"))
        font.set_family(self.style("famille"))
        text._rotation = self.style("angle")
        text._text = self.label()
        #debug(font.__dict__)



    def __init__(self, texte = "", x = None, y = None, **kw):
        if x == None:
            self.synchroniser_feuille()
            if self.__canvas__():
                xmin, xmax, ymin, ymax = self.__canvas__().fen()
                x = random.uniform(xmin,xmax)
                y = random.uniform(ymin,ymax)
                del xmin, xmax, ymin, ymax
            else:
                x = y = 0

        # on convertit l'entree en Variable
        # l'entree peut-etre un nombre, un(e longueur de) segment, un (rayon de) cercle
        x, y = self.nb2var(x), self.nb2var(y)
        
        Objet.__init__(self)
        self.style(**param.textes)
        if texte:   self.style(label = texte)
        self.enregistre(str, Variable, Variable)

    def _espace_vital(self):
        x, y = self.coordonnees()
        renderer = self.__representation__[0]._renderer
        if renderer is None:
            return (x, x, y, y)
        else:
            box = self.__representation__[0].get_window_extent()
            can = self.__canvas__()
            return (can.Xpix2coo(box.xmin()), can.Xpix2coo(box.xmax()), can.Ypix2coo(box.ymin()), can.Ypix2coo(box.ymax()))
    


    def _distance_inf(self, x, y, d):
        x1, x2, y1, y2 = self._espace_vital()
        can = self.__canvas__()
        #print can.Xcoo2pix(x1), can.Xcoo2pix(x2), can.Ycoo2pix(y1, -1), can.Ycoo2pix(y2, -1)
        return can.Xcoo2pix(x1) - d < x < can.Xcoo2pix(x2) + d and can.Ycoo2pix(y2, -1) - d < y < can.Ycoo2pix(y1, -1) + d
        
        """renderer = self.__representation__[0]._renderer
        if renderer is None:
            x0, y0 = self.position()
            return (x - x0)**2 + (y - y0)**2 < d**2
        else:
            box = self.__representation__[0].get_window_extent()
            return box.xmin() - d < x < box.xmax() + d and box.ymin() - d < y < box.ymax() + d"""

    #_call = _coordonnees

##########################################################################################

# Intersections

class Intersection_generique(Point_generique):
    """Une intersection generique.
    
    La classe mere des differents types d'intersections"""
    
    def __init__(self, objet1, objet2, **kw):
        Point_generique.__init__(self)
        self.enregistre(Objet, Objet)
        
        
    # L'intersection de 2 objets est parfois vide.
    # Un objet construit a partir d'un point d'intersection de 2 objets peut donc ne plus exister temporairement.
    # En particulier, il est important de verifier que l'objet existe avant de l'afficher.
        
#    def _affiche(self):
#        if self.coo():
#            self.__canvas__().Point(self.coo(), "b")
#            self.__canvas__().Texte(self.label(), self.coo() + array((3.*self.__canvas__().coeff(0), 2.*self.__canvas__().coeff(1))))





"""class Intersection_droites(Intersection_generique):"""
"""une intersection de droites.
    
    L'intersection de 2 droites"""
    
"""    def __init__(self, droite1, droite2, **kw):
        Intersection_generique.__init__(self, objet1 = droite1, objet2 = droite2)        
        self.enregistre(Droite, Droite)
       
    
    def _conditions_existence(self):
        a, b, c = self.droite1.equation()
        d, e, f = self.droite2.equation()
        self.__coeffs = a, b, c, d, e, f
        self.__determinant = a*e - b*d
        return [abs(self.__determinant) > param.tolerance]


    def _coordonnees(self):
        a, b, c, d, e, f = self.__coeffs
        x = (f*b - c*e)/self.__determinant
        y = (d*c - a*f)/self.__determinant
        return (x, y)"""


class Intersection_droites(Intersection_generique):
    """une intersection de droites.
    
    L'intersection de 2 droites, segments, demi-droites..."""
    
    def __init__(self, droite1, droite2, **kw):
        Intersection_generique.__init__(self, objet1 = droite1, objet2 = droite2)        
        self.enregistre((Droite, Segment, Demidroite), (Droite, Segment, Demidroite))
       
    
    def _conditions_existence(self):
        a, b, c = self.droite1.equation()
        d, e, f = self.droite2.equation()
        #self.__coeffs = a, b, c, d, e, f
        self.__determinant = a*e - b*d
        if abs(self.__determinant) <= param.tolerance:
            return [False]
        x = self.__x = (f*b - c*e)/self.__determinant
        y = self.__y = (d*c - a*f)/self.__determinant
        M = Point(x, y)
        return [isinstance(self.droite1, Droite) or M in self.droite1, isinstance(self.droite2, Droite) or M in self.droite2]


    def _coordonnees(self):
        return (self.__x, self.__y)



class Intersection_droite_cercle(Intersection_generique):
    """Une intersection d'une droite et d'un cercle.
    
    Un des deux points M et N d'intersection d'une droite et d'un cercle.
    Supposons qu'on le note M, et que la droite soit (AB).
    L'argument point_proche (optionnel) doit etre un des deux points de la droite (A par exemple).
    Il sert a indiquer "de quel cote" de la droite (AB) se trouve le point M.
    (Plus exactement, dans l'exemple choisi, MA + NB est minimal).
    """
    
    def __init__(self, droite, cercle, point_proche = None, **kw):
        # l'intersection de la droite et du cercle (si elle existe) est deux points (eventuellement confondus).
        # lorsque l'utilisateur cree deux fois deux suite un point d'intersection, ce ne doit pas etre le meme.
        self.synchroniser_feuille()
        if point_proche == None:
            if self.__feuille__:
                intersections = [objet for objet in self.__feuille__.objets() if isinstance(objet, Intersection_droite_cercle) and objet.droite is droite and objet.cercle is cercle]
                if intersections:
                    point_proche = intersections[0].point_lointain
                else:
                    point_proche = droite.point1
                del intersections
            else:
                point_proche = droite.point1
        
        self.point_lointain = (droite.point1 == point_proche) and droite.point2 or droite.point1
        Intersection_generique.__init__(self, objet1 = droite, objet2 = cercle)
        self.enregistre((Droite, Segment, Demidroite), (Cercle_generique, Arc_cercle), Point_generique)


    def _coordonnees(self, bipoint = False):
        if bipoint:
            return self.__sol1, self.__sol2
        return self.__sol1
        # on a dj calcul les coordonnes en cherchant les conditions d'existence
    
    
    
    def _conditions_existence(self):
        """Si bipoint = True, retourne les coordonnees des deux points d'intersection.
        (Dans l'ordre: celles du point, puis celles du 2eme point d'intersection).
        """
        # Trouver les conditions d'existence revient pratiquement  chercher les coordonnes.
        # Tout le calcul de coordonnes se fait donc ici, pour simplifier.
        
        a, b, c = self.droite.equation()
        d, e, f = self.cercle.equation()

        if b: # La droite n'est pas verticale
            m = -a/b; n = -c/b # equation de la droite : y = mx + n
            p = Polynome([m**2 + 1, 2*n*m + d + e*m, n**2 + e*n + f])

            if p.delta() > 0: # 2 points d'intersection.
                x1, x2 = p.solutions(False)
                M = Point(x1, m*x1+n); N = Point(x2, m*x2+n)
            elif p.delta() == 0: # 1 seul point
                x = p.solutions(False)[0]
                self.__sol1 = self.__sol2 = (x, m*x+n)
                M = Point(*self.__sol1)
                return [isinstance(self.droite, Droite) or M in self.droite, isinstance(self.cercle, Cercle_generique) or M in self.cercle]
            else:
                return [False] # intersection vide.

        else: # La droite est verticale
            x = -c/a
            p = Polynome([1, e, x**2 + d*x + f])

            if p.delta() > 0: # 2 points d'intersection.
                y1, y2 = p.solutions(False)
                M = Point(x, y1); N = Point(x, y2)
            elif p.delta() == 0:
                y = p.solutions(False)[0]
                self.__sol1 = self.__sol2 = (x, y)
                M = Point(*self.__sol1)
                return [isinstance(self.droite, Droite) or M in self.droite, isinstance(self.cercle, Cercle_generique) or M in self.cercle]
            else:
                return [False] # intersection vide.

        # La partie la plus complexe consiste a distinguer les deux points d'intersection a l'aide du "point proche".
        # Pour comprendre ce qui s'y passe, il est conseille de faire un dessin !
        
        A, B = self.point_proche, self.point_lointain
        u = Vecteur(A, B)
        if produit_scalaire(u, Vecteur(A, M)) < produit_scalaire(u, Vecteur(A, N)):
            self.__sol1, self.__sol2 = M.coordonnees(), N.coordonnees()
            return [isinstance(self.droite, Droite) or M in self.droite, isinstance(self.cercle, Cercle_generique) or M in self.cercle]
        self.__sol1, self.__sol2 = N.coordonnees(), M.coordonnees()
        return [isinstance(self.droite, Droite) or N in self.droite, isinstance(self.cercle, Cercle_generique) or N in self.cercle]
        
        
  
  
  

class Intersection_cercles(Intersection_generique):
    """Une intersection de cercles.
    
    Un des deux points M et N d'intersection des deux cercles.
    L'argument angle_positif permet de distinguer les deux points.
    Si on note A et B les centres respectifs des 2 cercles, l'angle oriente (AB>, AM>) doit etre de signe constant.
    """
    def __init__(self, cercle1, cercle2, angle_positif = None, **kw):
        # l'intersection de deux cercles (si elle existe) est deux points (eventuellement confondus).
        # lorsque l'utilisateur cree deux fois de suite un point d'intersection, ce ne doit pas etre le meme.
        self.synchroniser_feuille()
        if self.__feuille__:
            if angle_positif == None:
                intersections = [objet for objet in self.__feuille__.objets() if isinstance(objet, Intersection_cercles) and objet.cercle1 is cercle1 and objet.cercle2 is cercle2]
                if intersections:
                    angle_positif = intersections[0].angle == -1
                else:
                    angle_positif = True
                del intersections
        else:
            angle_positif = True
                
        angle = angle_positif and 1 or -1
        Intersection_generique.__init__(self, objet1 = cercle1, objet2 = cercle2)
        self.enregistre((Cercle_generique, Arc_cercle), (Cercle_generique, Arc_cercle), bool)


    def _conditions_existence(self):
        a, b, c = self.cercle1.equation()
        d, e, f = self.cercle2.equation()

        if (a, b, c) == (d, e, f):
            return None

        # (a-d)x + (b-e)y + (c-f) = 0 -> equation d'une droite.
        d = Droite_equation((a - d, b - e, c - f))
        M = Intersection_droite_cercle(d, self.cercle1, d.point1) # 1er point d'intersection
        N = Intersection_droite_cercle(d, self.cercle1, d.point2) # 2eme point d'intersection
        
        # La encore, la partie la plus complexe consiste a distinguer les deux points d'intersection, a l'aide cette fois du signe de l'angle.
        # Pour comprendre ce qui s'y passe, il est conseille de faire un dessin !
        A, B = self.cercle1.centre, self.cercle2.centre
        AB = Vecteur(A, B); AM = Vecteur(A, M)
        a = Angle_oriente(AB, AM).val()
        signe_a = a/abs(a)
        #print "signe de a : ", signe_a
        #print "signe choisi : ", self.angle
        #print M.coordonnees()
        #print N.coordonnees()
        
        if signe_a == self.angle:
            self.__sol1 = M.coordonnees()
            self.__sol2 = N.coordonnees()
            return [M.existe(), isinstance(self.cercle1, Cercle_generique) or M in self.cercle1, isinstance(self.cercle2, Cercle_generique) or M in self.cercle2]
        else:
            self.__sol1 = N.coordonnees()
            self.__sol2 = M.coordonnees()
            return [N.existe(), isinstance(self.cercle1, Cercle_generique) or N in self.cercle1, isinstance(self.cercle2, Cercle_generique) or N in self.cercle2]




    def _coordonnees(self, bipoint = False):
        """Si bipoint = True, retourne les coordonnees des deux points d'intersection.
        (Dans l'ordre: celles du point, puis celles du 2eme point d'intersection).
        """
        
        if bipoint:
            return self.__sol1, self.__sol2
        return self.__sol1
        # on a dj calcul les coordonnes en cherchant les conditions d'existence



"""        # La encore, la partie la plus complexe consiste a distinguer les deux points d'intersection, a l'aide cette fois du signe de l'angle.
        # Pour comprendre ce qui s'y passe, il est conseille de faire un dessin !
        M, N = self.__M, self.__N
        A, B = self.cercle1.centre, self.cercle2.centre
        AB = Vecteur(A, B); AM = Vecteur(A, M)
        a = Angle_oriente(AB, AM).val()
        signe_a = a/abs(a)
        #print "signe de a : ", signe_a
        #print "signe choisi : ", self.angle
        #print M.coordonnees()
        #print N.coordonnees()
        
        if bipoint:
            if signe_a == self.angle:
                return M.coordonnees(), N.coordonnees()
            else:
                return N.coordonnees(), M.coordonnees()
        return (signe_a == self.angle) and M.coordonnees() or N.coordonnees()"""


#############################################################################################
#############################################################################################
#############################################################################################
# <SUPPRIMER>

class OldIntersection_cercles(Intersection_generique):
    """Une intersection de cercles.
    
    Un des deux points M et N d'intersection des deux cercles.
    L'argument angle_positif permet de distinguer les deux points.
    Si on note A et B les centres respectifs des 2 cercles, l'angle oriente (AB>, AM>) doit etre de signe constant.
    """
    def __init__(self, cercle1, cercle2, angle_positif = None, **kw):
        # l'intersection de deux cercles (si elle existe) est deux points (eventuellement confondus).
        # lorsque l'utilisateur cree deux fois de suite un point d'intersection, ce ne doit pas etre le meme.
        self.synchroniser_feuille()
        if self.__feuille__:
            if angle_positif == None:
                intersections = [objet for objet in self.__feuille__.objets() if isinstance(objet, Intersection_cercles) and objet.cercle1 is cercle1 and objet.cercle2 is cercle2]
                if intersections:
                    angle_positif = intersections[0].angle == -1
                else:
                    angle_positif = True
                del intersections
        else:
            angle_positif = True
                
        angle = angle_positif and 1 or -1
        Intersection_generique.__init__(self, objet1 = cercle1, objet2 = cercle2)
        self.enregistre(Cercle_generique, Cercle_generique, bool)


    def _conditions_existence(self):
        a, b, c = self.cercle1.equation()
        d, e, f = self.cercle2.equation()

        if (a, b, c) == (d, e, f):
            return None

        # (a-d)x + (b-e)y + (c-f) = 0 -> equation d'une droite.
        d = Droite_equation((a - d, b - e, c - f))
        self.__M = Intersection_droite_cercle(d, self.cercle1, d.point1) # 1er point d'intersection
        self.__N = Intersection_droite_cercle(d, self.cercle1, d.point2) # 2eme point d'intersection
        
        return [self.__M.existe()]



    def _coordonnees(self, bipoint = False):
        """Si bipoint = True, retourne les coordonnees des deux points d'intersection.
        (Dans l'ordre: celles du point, puis celles du 2eme point d'intersection).
        """
        
        # La encore, la partie la plus complexe consiste a distinguer les deux points d'intersection, a l'aide cette fois du signe de l'angle.
        # Pour comprendre ce qui s'y passe, il est conseille de faire un dessin !
        M, N = self.__M, self.__N
        A, B = self.cercle1.centre, self.cercle2.centre
        AB = Vecteur(A, B); AM = Vecteur(A, M)
        a = Angle_oriente(AB, AM).val()
        signe_a = a/abs(a)
        #print "signe de a : ", signe_a
        #print "signe choisi : ", self.angle
        #print M.coordonnees()
        #print N.coordonnees()
        
        if bipoint:
            if signe_a == self.angle:
                return M.coordonnees(), N.coordonnees()
            else:
                return N.coordonnees(), M.coordonnees()
        return (signe_a == self.angle) and M.coordonnees() or N.coordonnees()



class Intersection_segments(Intersection_droites):
    """une intersection de segments.
    
    L'intersection de 2 segments"""
    
    def __init__(self, segment1, segment2, **kw):
        Intersection_droites.__init__(self, segment1, segment2)        
        self.enregistre(Segment, Segment)
       
    
    def _conditions_existence(self):
        if False not in Intersection_droites._conditions_existence(self): # veiller a ce que _conditions_existence soit appele AVANT _coordonnees (_coordonnees reutilise certains des calculs de _conditions_existence)
            xy = Intersection_droites._coordonnees(self)
            M = Point(*xy)
            if M in self.segment1 and M in self.segment2: # on verifie que l'intersection appartient au 2 segments.
                self.__couple = xy
                return [True]
        return [False]
        
    
    def _coordonnees(self):
        return self.__couple
                
                
                
                
class Intersection_segment_droite(Intersection_droites):
    """une intersection d'un segment et d'une droite.
    
    L'intersection d'un segment et d'une droite."""
    
    def __init__(self, segment, droite, **kw):
        Intersection_droites.__init__(self, segment, droite)        
        self.enregistre(Segment, Droite)
       
    
    def _conditions_existence(self):
        if False not in Intersection_droites._conditions_existence(self): # veiller a ce que _conditions_existence soit appele AVANT _coordonnees (_coordonnees reutilise certains des calculs de _conditions_existence)
            xy = Intersection_droites._coordonnees(self)
            if Point(*xy) in self.segment: # on verifie que l'intersection appartient au segment.
                self.__couple = xy
                return [True]
        return [False]        
    
    def _coordonnees(self):
        return self.__couple



class Intersection_droite_segment(Intersection_segment_droite):
    """une intersection d'une droite et d'un segment.
    
    Alias pour l'intersection d'un segment et d'une droite, mais les arguments sont inverss."""
    
    def __init__(self, droite, segment, **kw):
        Intersection_segment_droite.__init__(self, segment, droite)        
        self.enregistre(Droite, Segment)




class Intersection_segment_cercle(Intersection_droite_cercle):
    """une intersection d'un segment et d'un cercle.
    
    Un des deux points M et N d'intersection d'un segment et d'un cercle.
    Supposons qu'on le note M, et que le segment soit porte par la droite (AB).
    L'argument point_proche (optionnel) doit etre un des deux points de la droite (A par exemple).
    Il sert a indiquer "de quel cote" de la droite (AB) se trouve le point M.
    (Plus exactement, dans l'exemple choisi, MA + NB est minimal).
    """
    
    def __init__(self, segment, cercle, point_proche = None, **kw):
        # l'intersection de la droite et du cercle (si elle existe) est deux points (eventuellement confondus).
        # lorsque l'utilisateur cree deux fois deux suite un point d'intersection, ce ne doit pas etre le meme.
        self.synchroniser_feuille()
        if point_proche == None:
            if self.__feuille__:
                intersections = [objet for objet in self.__feuille__.objets() if isinstance(objet, Intersection_segment_cercle) and objet.segment is segment and objet.cercle is cercle]
                if intersections:
                    point_proche = intersections[0].point_lointain
                else:
                    if Intersection_segment_cercle(segment, cercle, segment.point1).existe():
                        point_proche = segment.point1
                    else:
                        point_proche = segment.point2
                del intersections
            else:
                if Intersection_segment_cercle(segment, cercle, segment.point1).existe():
                    point_proche = segment.point1
                else:
                    point_proche = segment.point2
        
        self.point_lointain = (segment.point1 == point_proche) and segment.point2 or segment.point1
        Intersection_generique.__init__(self, objet1 = segment, objet2 = cercle)
        self.enregistre(Segment, Cercle_generique, Point_generique)


    def _coordonnees(self, bipoint = False):
        if bipoint:
            return self.__sol1, self.__sol2
        return self.__sol1
        # on a dj calcul les coordonnes en cherchant les conditions d'existence
    
    
    
    def _conditions_existence(self):
        """Si bipoint = True, retourne les coordonnees des deux points d'intersection.
        (Dans l'ordre: celles du point, puis celles du 2eme point d'intersection).
        """
        # Trouver les conditions d'existence revient pratiquement  chercher les coordonnes.
        # Tout le calcul de coordonnes se fait donc ici, pour simplifier.
        
        a, b, c = self.segment.equation()
        d, e, f = self.cercle.equation()

        if b: # La droite n'est pas verticale
            m = -a/b; n = -c/b # equation de la droite : y = mx + n
            p = Polynome([m**2 + 1, 2*n*m + d + e*m, n**2 + e*n + f])

            if p.delta() > 0: # 2 points d'intersection.
                x1, x2 = p.solutions(False)
                M = Point(x1, m*x1+n); N = Point(x2, m*x2+n)
            elif p.delta() == 0: # 1 seul point
                x = p.solutions(False)[0]
                self.__sol1 = self.__sol2 = (x, m*x+n)
                return [Point(**self.__sol1) in self.segment]
            else:
                return [False] # intersection vide.

        else: # La droite est verticale
            x = -c/a
            p = Polynome([1, e, x**2 + d*x + f])

            if p.delta() > 0: # 2 points d'intersection.
                y1, y2 = p.solutions(False)
                M = Point(x, y1); N = Point(x, y2)
            elif p.delta() == 0:
                y = p.solutions(False)[0]
                self.__sol1 = self.__sol2 = (x, y)
                return [Point(**self.__sol1) in self.segment]
            else:
                return [False] # intersection vide.

        # La partie la plus complexe consiste a distinguer les deux points d'intersection a l'aide du "point proche".
        # Pour comprendre ce qui s'y passe, il est conseille de faire un dessin !
        
        A, B = self.point_proche, self.point_lointain
        u = Vecteur(A, B)
        if produit_scalaire(u, Vecteur(A, M)) < produit_scalaire(u, Vecteur(A, N)):
            self.__sol1, self.__sol2 = M.coordonnees(), N.coordonnees()
            return [M in self.segment]
        self.__sol1, self.__sol2 = N.coordonnees(), M.coordonnees()
        return [N in self.segment]


"""    def __init__(self, segment, cercle, point_proche = None, **kw):
        Intersection_droite_cercle.__init__(self, segment, cercle, point_proche)
        point_proche = self.point_proche # valide les tranformations effectuees par Intersection_droite_cercle.__init__
        self.enregistre(Segment, Cercle_generique, Point_generique)


    def _conditions_existence(self):
        if False not in Intersection_droite_cercle._conditions_existence():
            xy = Intersection_droite_cercle._coordonnees(self, bipoint = True)
        if bipoint:
            xy1, xy2 = xy
            return xy1 and Point(*xy1) in self.segment and xy1 or None, xy2 and Point(*xy2) in self.segment and xy2 or None
        else: 
            if xy is not None:
                if Point(*xy) in self.segment: # on verifie que l'intersection appartient au segment.
                    return xy


    
    
    def _coordonnees(self, bipoint = False):
        xy = Intersection_droite_cercle._coordonnees(self, bipoint)
        if bipoint:
            xy1, xy2 = xy
            return xy1 and Point(*xy1) in self.segment and xy1 or None, xy2 and Point(*xy2) in self.segment and xy2 or None
        else: 
            if xy is not None:
                if Point(*xy) in self.segment: # on verifie que l'intersection appartient au segment.
                    return xy"""


# </SUPPRIMER>
#############################################################################################
#############################################################################################
#############################################################################################




##########################################################################################

# Glisseurs


class Glisseur_generique(Point_generique):
    """Un glisseur generique.
    
    Classe mere des differents types de glisseurs"""

    def __init__(self, objet, k = 0, **kw):
        if isinstance(k, (list, tuple)):
            pt = k
            k = self.nb2var(0)
        else:
            pt = None
            k = self.nb2var(k)
        Point_generique.__init__(self)
        self.style(**param.points_deplacables)
        self.enregistre(Objet, Variable)
        if pt is not None:
            self._coordonnees(pt[0], pt[1], False)

    def parametre(self, coeff = None):
        """Cette fonction devrait surtout etre utilisee en lecture.
        Alias plus 'expressif' de self.k(coeff)."""
        if coeff <> None: self.k(coeff)
        return self.k.val()


    def coo(self):
        return (1 - self.k.val())*self.objet.point1.coo() + self.k.val()*self.objet.point2.coo()
        
        
        

class Glisseur_droite(Glisseur_generique):
    """Un point sur une droite.
     
    Point pouvant 'glisser' sur une droite. Parametre est un coefficient barycentrique.
    Si A et B sont les deux points de reference de la droite, k le coefficient,
    les coordonnees du glisseur M seront determinees par la formule M = kA + (1-k)B."""

    def __init__(self, droite, k = None, **kw):
        if k == None:
            k = normalvariate(0.5, 0.5)
        # k = self.nb2var(k)
        Glisseur_generique.__init__(self, objet = droite, k = k)
        del k # k a ete enregistre apres modification dans Glisseur_generique (evitons de perdre les modifications !).
        # self.k = k # sinon, k a ete remplace par une copie (et l'enregistrement ne marche pas) 
        self.enregistre(Droite, Variable)
                
    
    def _coordonnees(self, x = None, y = None, auto_refresh = True):
        if x <> None:
            # permet d'utiliser le meme code pour Glisseur_droite, Glisseur_segment, Glisseur_demidroite
            if isinstance(self, Glisseur_segment):
                Projete = Projete_segment
            elif isinstance(self, Glisseur_demidroite):
                Projete = Projete_demidroite
            else:
                Projete = Projete_droite 
            M = Projete(Point(x, y), self.objet)
            A = self.objet.point1; B = self.objet.point2
            self.k(produit_scalaire(Vecteur(A, M), Vecteur(A, B))/Segment(A, B)._longueur()**2, auto_refresh)
        return self.coo()


        

class Glisseur_segment(Glisseur_droite):
    """Un point sur un segment.
    
    Point pouvant 'glisser' sur un segment. Parametre est un coefficient barycentrique.
    Si A et B sont les deux points de reference de la droite, k le coefficient,
    les coordonnees du glisseur M seront determinees par la formule M = kA + (1-k)B."""

    def __init__(self, segment, k = None, **kw):
        Glisseur_droite.__init__(self, droite = segment, k = k)
        del k # k a ete enregistre apres modification dans Glisseur_droite (evitons de perdre les modifications !).
        self.enregistre(Segment, Variable)


        

class Glisseur_demidroite(Glisseur_droite):
    """Un point sur un segment.
    
    Point pouvant 'glisser' sur un segment. Parametre est un coefficient barycentrique.
    Si A et B sont les deux points de reference de la droite, k le coefficient,
    les coordonnees du glisseur M seront determinees par la formule M = kA + (1-k)B."""

    def __init__(self, demidroite, k = None, **kw):
        Glisseur_droite.__init__(self, droite = demidroite, k = k)
        del k # k a ete enregistre apres modification dans Glisseur_droite (evitons de perdre les modifications !).
        self.enregistre(Demidroite, Variable)



class Glisseur_cercle(Glisseur_generique):
    """Un point sur un cercle.
    
    Point pouvant 'glisser' sur un cercle. Parametre est un angle.
    Si O et r sont les centres et rayons du cercle, a le parametre,
    les coordonnees du glisseur M seront determinees par la formule M = O + r*(cos(a), sin(a))."""

    def __init__(self, cercle, k = None, **kw):
        if k == None:
            k = 2*pi*rand()
        # k = self.nb2var(k)
        Glisseur_generique.__init__(self, objet = cercle, k = k)
        del k # k a ete enregistre apres modification dans Glisseur_generique (evitons de perdre les modifications !).
        # self.k = k # sinon, k a ete remplace par une copie (et l'enregistrement ne marche pas) 
        self.enregistre(Cercle_generique, Variable)
        
        
    def coo(self):
        return self.objet.centre.coo() + self.objet.rayon()*array((cos(self.parametre()), sin(self.parametre())))
    
    
    def _coordonnees(self, x = None, y = None, auto_refresh = True):
        if x <> None:
            O = self.objet.centre; M = Point(x, y)
            self.k(Angle_oriente(Vecteur_libre(1, 0),Vecteur(O, M)).val(), auto_refresh)
        return self.coo()
        


class Glisseur_arc_cercle(Glisseur_cercle):
    """Un point sur un cercle.
    
    Point pouvant 'glisser' sur un arc de cercle. Parametre est un angle.
    Si O et r sont les centres et rayons du cercle, a le parametre,
    les coordonnees du glisseur M seront determinees par la formule M = O + r*(cos(a), sin(a))."""

    def __init__(self, arc, k = None, **kw):
        if k == None:
            a = arc.angle1.val()
            b = arc.angle2.val()
            if b < a:
                b += 2*pi
            k = a + (b-a)*rand()
        # k = self.nb2var(k)
        Glisseur_cercle.__init__(self, objet = arc, k = k)
        del k # k a ete enregistre apres modification dans Glisseur_generique (evitons de perdre les modifications !).
        # self.k = k # sinon, k a ete remplace par une copie (et l'enregistrement ne marche pas) 
        self.enregistre(Arc_cercle, Variable)
        
            
    
    def _coordonnees(self, x = None, y = None, auto_refresh = True):
        if x <> None:
            O = self.objet.centre; M = Point(x, y)
            a = self.objet.angle1.val()
            b = self.objet.angle2.val()
            if b < a:
                b += 2*pi
            c = Angle_oriente(Vecteur_libre(1, 0),Vecteur(O, M)).val()
            if c < a:
                c += 2*pi
            # La mesure d'angle c est donc dans l'intervalle [a; a+2*pi[
            if c > b: # c n'appartient pas a [a;b] (donc M est en dehors de l'arc de cercle)
                if c - b > 2*pi + a - c: # c est plus proche de a+2*pi
                    c = a
                else:   # c est plus proche de b
                    c = b
            self.k(c, auto_refresh)
        return self.coo()




##########################################################################################

# Polygones

class Polygone(Objet):
    """Un polygone.
    
    Classe mere de tous les polygones."""
    
    def __init__(self, *points, **kw):
        if self.classe() == "Polygone" and len(points) == 3:
            self.__transformer_en_triangle(*points, **kw)
        else:
            if isinstance(points[0], (list, tuple)):
                points = tuple(points[0])
            sommets = points # alias
            centre = centre_gravite = Barycentre(points)
            Objet.__init__(self)
            self.style(**param.polygones)
            self.enregistre([Point_generique])
            #self.classifier()
            self.creer_cotes()
        
        
    def creer_cotes(self):
        if hasattr(self, "hierarchie") and not hasattr(self, "cotes"):
        # un triangle subit 2 initialisations
            self.cotes = [Cote(self, i) for i in range(len(self.points))]
            if self.__feuille__:
                for cote in self.cotes:
                    #nom = self.__feuille__.nom_alea("c")
                    #print nom
                    setattr(self.__feuille__.objets, self.__feuille__.nom_alea("c"), cote)
    
    
    def _affiche(self):
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().fill(), self.__canvas__().plot(zorder = 1)]
        
        fill = self.__representation__[0]
        plot = self.__representation__[1]
        points = self.points + (self.points[0],)
        xy = [pt.coordonnees() for pt in points]
        x, y = zip(*xy)
        fill.xy = xy
        plot._x = x; plot._y = y
        fill._alpha = self.style("alpha")
        fill._edgecolor = fill._facecolor = plot._color = self.style("couleur")
        plot._linestyle = fill._linestyle = self.style("style")
        plot._linewidth = fill._linewidth = self.style("epaisseur")

    #def centre_gravite(self):
    #    return Barycentre(self.points).coo()

    #centre = centre_gravite
    
    def __transformer_en_triangle(self, *points, **kw):
        self.__class__ = Triangle
        Triangle.__init__(self, *points, **kw)
            
    def _distance_inf(self, x, y, d):
        #points = self.points + (self.points[0],)
        xy = [pt.position() for pt in self.points]
        #return True in [cote._distance_inf(x, y, d) for cote in self.cotes] or point_dans_polygone((x,y), xy)
        return point_dans_polygone((x,y), xy)

    def __contains__(self, M):
        if not isinstance(M, Point_generique):
            return False
        xy = [pt.coordonnees() for pt in self.points]
        return True in [(M in cote) for cote in self.cotes] or point_dans_polygone(M.coordonnees(), xy)
        
    def aire(self):
        """D'apres David Chandler, Area of a general polygon.
        http://www.davidchandler.com/AreaOfAGeneralPolygon.pdf"""

        if self.existe():
            xy = [pt.coordonnees() for pt in (self.points + (self.points[0],))]
            s1 = s2 = 0
            for i in range(len(self.points)):
                s1 += xy[i][0]*xy[i+1][1]
                s2 += xy[i][1]*xy[i+1][0]
            return .5*abs(s1-s2)
        return None
        
    def perimetre(self):
        return builtin.sum(cote.longueur() for cote in self.cotes)
        
    def _espace_vital(self):        
        x1 = min(pt.abscisse() for pt in self.points)
        x2 = max(pt.abscisse() for pt in self.points)
        y1 = min(pt.ordonnee() for pt in self.points)
        y2 = max(pt.ordonnee() for pt in self.points)
        return (x1, x2, y1, y2)
        
        

class Triangle(Polygone):
    """Un triangle."""
    
    def __init__(self, point1, point2, point3, **kw):
        Polygone.__init__(self, point1, point2, point3)
        d1 = Perpendiculaire(Droite(point1, point2), point3)
        d2 = Perpendiculaire(Droite(point2, point3), point1)
        orthocentre = Intersection_droites(d1, d2)
        d1 = Mediatrice(point1, point2)
        d2 = Mediatrice(point2, point3)
        centre_cercle_circonscrit = Intersection_droites(d1, d2)
        d1 = Bissectrice(point1, point2, point3)
        d2 = Bissectrice(point3, point1, point2)
        centre_cercle_inscrit = Intersection_droites(d1, d2)
        del d1, d2
        self.enregistre(Point_generique, Point_generique, Point_generique)
        self.creer_cotes() # les cotes doivent etre crees quand l'objet est acheve (donc apres enregistrement).

    #def orthocentre(self):
    #    d1 = Perpendiculaire(Droite(self.point1, self.point2), self.point3)
    #    d2 = Perpendiculaire(Droite(self.point2, self.point3), self.point1)
    #    return Intersection_droites(d1, d2).coo()
        
    #def centre_cercle_circonscrit(self):
    #    d1 = Mediatrice(self.point1, self.point2)
    #    d2 = Mediatrice(self.point2, self.point3)
    #    return Intersection_droites(d1, d2).coo()
    
    #def centre_cercle_inscrit(self):
    #    d1 = Bissectrice(self.point1, self.point2, self.point3)
    #    d2 = Bissectrice(self.point3, self.point1, self.point2)
    #    return Intersection_droites(d1, d2).coo()



class Cote(Segment):
    """Un cote.
    
    Un cote d'un polygone, reliant le point numero 'n' au point numero 'n + 1'.
    Note: n commence a 0.
    L'objet est cree automatiquement lors de la creation du polygone.
    De plus, si l'objet est supprime, le polygone est automatiquement supprime."""

    def __init__(self, polygone, n):
        Segment.__init__(self, polygone.points[n], polygone.points[(n + 1)%len(polygone.points)])
        self.style(**polygone.style())
        self.enregistre(Polygone, int)
        
    def supprimer(self):
        self.polygone.supprimer()

# Classes a rajouter :
# polygones reguliers, triangles equilateraux
        